package de.mdelab.mlsdm.interpreter;

import org.eclipse.emf.ecore.EClassifier;
import org.eclipse.emf.ecore.EStructuralFeature;

import de.mdelab.expressions.interpreter.core.ExpressionInterpreterManager;
import de.mdelab.mlexpressions.MLExpression;
import de.mdelab.mlsdm.Activity;
import de.mdelab.mlsdm.ActivityEdge;
import de.mdelab.mlsdm.ActivityNode;
import de.mdelab.mlsdm.interpreter.facade.MLSDMMetamodelFacadeFactory;
import de.mdelab.mlsdm.interpreter.notifications.MLSDMNotificationEmitter;
import de.mdelab.mlsdm.interpreter.patternMatcher.patternPartBased.MLSDMPatternPartBasedMatcher;
import de.mdelab.mlstorypatterns.AbstractStoryPatternLink;
import de.mdelab.mlstorypatterns.AbstractStoryPatternObject;
import de.mdelab.mlstorypatterns.StoryPattern;
import de.mdelab.sdm.interpreter.core.SDMException;
import de.mdelab.sdm.interpreter.core.SDMInterpreter;
import de.mdelab.sdm.interpreter.core.facade.MetamodelFacadeFactory;
import de.mdelab.sdm.interpreter.core.notifications.NotificationEmitter;
import de.mdelab.sdm.interpreter.core.patternmatcher.MatchApplier;
import de.mdelab.sdm.interpreter.core.patternmatcher.StoryPatternMatcher;
import de.mdelab.sdm.interpreter.core.variables.NotifierVariablesScope;

/**
 * 
 * @author Stephan Hildebrandt
 * 
 */
public class MLSDMInterpreter
		extends
		SDMInterpreter<Activity, ActivityNode, ActivityEdge, StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> {
	
	public MLSDMInterpreter( 	final ExpressionInterpreterManager<EClassifier, EStructuralFeature, MLExpression> expressionInterpreterManager )	{
		this( expressionInterpreterManager, new MLSDMNotificationEmitter() );
	}

	public MLSDMInterpreter( 	final ExpressionInterpreterManager<EClassifier, EStructuralFeature, MLExpression> expressionInterpreterManager,
								final NotificationEmitter<Activity, ActivityNode, ActivityEdge, StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> notificationEmitter )	{
		this( 	expressionInterpreterManager,
				notificationEmitter, //expressionInterpreterManager.getNotificationEmitter(),
				MLSDMMetamodelFacadeFactory.INSTANCE );
	}

	public MLSDMInterpreter(	final ExpressionInterpreterManager<EClassifier, EStructuralFeature, MLExpression> expressionInterpreterManager,
								final MetamodelFacadeFactory<Activity, ActivityNode, ActivityEdge, StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> facadeFactory )	{
		this( expressionInterpreterManager, new MLSDMNotificationEmitter(), facadeFactory );
	}

	public MLSDMInterpreter(	final ExpressionInterpreterManager<EClassifier, EStructuralFeature, MLExpression> expressionInterpreterManager,
								final NotificationEmitter<Activity, ActivityNode, ActivityEdge, StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> notificationEmitter,
								final MetamodelFacadeFactory<Activity, ActivityNode, ActivityEdge, StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> facadeFactory )	{
		super( facadeFactory, expressionInterpreterManager, notificationEmitter );
	}

	@Override
	protected StoryPatternMatcher<Activity, ActivityNode, ActivityEdge, StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> createStoryPatternMatcher(
			final StoryPattern storyPattern,
			final NotifierVariablesScope<Activity, ActivityNode, ActivityEdge, StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> variablesScope)
	throws SDMException	{
		return new MLSDMPatternPartBasedMatcher(	storyPattern,
													variablesScope,
													getFacadeFactory(),
													getExpressionInterpreterManager(),
													getNotificationEmitter() );
	}

	@Override
	protected MatchApplier<Activity, ActivityNode, ActivityEdge, StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> createMatchApplier(
			StoryPattern storyPattern) {
		return new MLSDMMatchApplier(storyPattern, getFacadeFactory(), getExpressionInterpreterManager(), getNotificationEmitter());
	}
	
	@Override
	protected ActivityNode executeCustomNode(
			final ActivityNode node,
			final NotifierVariablesScope<Activity, ActivityNode, ActivityEdge, StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> variablesScope)
	{
		throw new UnsupportedOperationException();
	}
	
	@Override
	public MLSDMNotificationEmitter getNotificationEmitter() {
		return (MLSDMNotificationEmitter) super.getNotificationEmitter();
	}
}
