package de.mdelab.mlsdm.interpreter;

import java.util.Collection;
import java.util.HashMap;
import java.util.Map;
import java.util.Map.Entry;

import org.eclipse.emf.common.util.EMap;
import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.EClassifier;
import org.eclipse.emf.ecore.EFactory;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EStructuralFeature;

import de.mdelab.expressions.interpreter.core.ExpressionsInterpreterException;
import de.mdelab.expressions.interpreter.core.ExpressionInterpreterManager;
import de.mdelab.expressions.interpreter.core.variables.Variable;
import de.mdelab.mlexpressions.MLExpression;
import de.mdelab.mlsdm.Activity;
import de.mdelab.mlsdm.ActivityEdge;
import de.mdelab.mlsdm.ActivityNode;
import de.mdelab.mlstorypatterns.AbstractStoryPatternLink;
import de.mdelab.mlstorypatterns.AbstractStoryPatternObject;
import de.mdelab.mlstorypatterns.AttributeAssignment;
import de.mdelab.mlstorypatterns.ContainmentLink;
import de.mdelab.mlstorypatterns.MapEntryLink;
import de.mdelab.mlstorypatterns.MlstorypatternsPackage;
import de.mdelab.mlstorypatterns.ModifierEnum;
import de.mdelab.mlstorypatterns.StoryPattern;
import de.mdelab.mlstorypatterns.StoryPatternLink;
import de.mdelab.mlstorypatterns.StoryPatternObject;
import de.mdelab.sdm.interpreter.core.SDMException;
import de.mdelab.sdm.interpreter.core.facade.IInstanceFacade;
import de.mdelab.sdm.interpreter.core.facade.IStoryPatternObjectFacade;
import de.mdelab.sdm.interpreter.core.facade.MetamodelFacadeFactory;
import de.mdelab.sdm.interpreter.core.notifications.NotificationEmitter;
import de.mdelab.sdm.interpreter.core.patternmatcher.MatchApplier;
import de.mdelab.sdm.interpreter.core.variables.NotifierVariablesScope;

public class MLSDMMatchApplier
		extends
			MatchApplier<Activity, ActivityNode, ActivityEdge, StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> {

	protected ExpressionInterpreterManager<EClassifier, EStructuralFeature, MLExpression> expressionInterpreterManager;
	protected IStoryPatternObjectFacade<AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> spoFacade;
	protected IInstanceFacade<AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> instanceFacade;

	public MLSDMMatchApplier(StoryPattern storyPattern,
			MetamodelFacadeFactory<Activity, ActivityNode, ActivityEdge, StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> facadeFactory,
			ExpressionInterpreterManager<EClassifier, EStructuralFeature, MLExpression> expressionInterpreterManager,
			NotificationEmitter<Activity, ActivityNode, ActivityEdge, StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> notificationEmitter) {
		super(storyPattern, notificationEmitter);
		
		this.expressionInterpreterManager = expressionInterpreterManager;

		this.spoFacade = facadeFactory.getStoryPatternObjectFacade();
		this.instanceFacade = facadeFactory.getInstanceFacade();
	}

	@Override
	public void applyToEncodedMatch(
			NotifierVariablesScope<Activity, ActivityNode, ActivityEdge, StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> matchScope) throws SDMException {
		Map<AttributeAssignment, Object> assignmentValues = computeAssignmentValues(matchScope);
		
		Map<AbstractStoryPatternObject, EObject> destroyedObjects = destroyObjects(matchScope);
		destroyLinks(matchScope, destroyedObjects);
		createObjects(matchScope);
		createLinks(matchScope);
		
		assignAttributeValues(assignmentValues, matchScope);
	}

	private Map<AttributeAssignment, Object> computeAssignmentValues(
			NotifierVariablesScope<Activity, ActivityNode, ActivityEdge, StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> matchScope)
	throws SDMException {
		Map<AttributeAssignment, Object> assignmentValues = new HashMap<AttributeAssignment, Object>();
		for(AbstractStoryPatternObject spo:storyPattern.getStoryPatternObjects()) {
			if(spo.eClass() == MlstorypatternsPackage.Literals.STORY_PATTERN_OBJECT) {
				Variable<EClassifier> var = matchScope.getVariable(spo.getName());
				if(var == null && spo.getModifier() != ModifierEnum.CREATE) {
					continue;
				}
				
				try {
					if(var != null) {
						Object contextInstance = var.getValue();
						for(AttributeAssignment assignment:((StoryPatternObject) spo).getAttributeAssignments()) {
							Object value = expressionInterpreterManager.evaluateExpression(assignment.getAssignmentExpression(), spo.getType(), contextInstance, matchScope).getValue();
							assignmentValues.put(assignment, value);
						}
					}
					else {
						for(AttributeAssignment assignment:((StoryPatternObject) spo).getAttributeAssignments()) {
							Object value = expressionInterpreterManager.evaluateExpression(assignment.getAssignmentExpression(), null, null, matchScope).getValue();
							assignmentValues.put(assignment, value);
						}
					}
				}
				catch ( final ExpressionsInterpreterException ex ) {
					throw new SDMException( ex );
				}
			}
		}
		return assignmentValues;
	}

	private Map<AbstractStoryPatternObject, EObject> destroyObjects(NotifierVariablesScope<Activity, ActivityNode, ActivityEdge, StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> matchScope) {
		Map<AbstractStoryPatternObject, EObject> destroyedObjects = new HashMap<AbstractStoryPatternObject, EObject>();
		for(AbstractStoryPatternObject spo:storyPattern.getStoryPatternObjects()) {
			if(spo.getModifier() == ModifierEnum.DESTROY) {
				EObject destroyedObject = destroyObject(spo, matchScope);
				if(destroyedObject != null) {
					destroyedObjects.put(spo, destroyedObject);
				}
			}
		}
		return destroyedObjects;
	}

	protected EObject destroyObject(AbstractStoryPatternObject spo, NotifierVariablesScope<Activity, ActivityNode, ActivityEdge, StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> matchScope) {
		EObject destroyedObject = null;
		String name = spoFacade.getName(spo);
		Variable<EClassifier> var = matchScope.getVariable(name);
		if(var != null) {
			matchScope.deleteVariable(name);
			destroyedObject = (EObject) var.getValue();
			instanceFacade.delete(destroyedObject);
			
			getNotificationEmitter().instanceObjectDestroyed(spo, destroyedObject, matchScope, this);
		}
		return destroyedObject;
	}

	private void destroyLinks(NotifierVariablesScope<Activity, ActivityNode, ActivityEdge, StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> matchScope, Map<AbstractStoryPatternObject, EObject> destroyedObjects) {
		for(AbstractStoryPatternLink link:storyPattern.getStoryPatternLinks()) {
			if(link.getModifier() == ModifierEnum.DESTROY) {
				destroyLink(link, matchScope, destroyedObjects);
			}
		}
	}

	private void destroyLink(AbstractStoryPatternLink link,
			NotifierVariablesScope<Activity, ActivityNode, ActivityEdge, StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> matchScope, Map<AbstractStoryPatternObject, EObject> destroyedObjects) {
		if(link.eClass() == MlstorypatternsPackage.Literals.STORY_PATTERN_LINK) {
			destroyStoryPatternLink((StoryPatternLink) link, matchScope, destroyedObjects);
		}
		else if(link.eClass() == MlstorypatternsPackage.Literals.CONTAINMENT_LINK) {
			destroyContainmentLink((ContainmentLink) link, matchScope, destroyedObjects);
		}
		else if(link.eClass() == MlstorypatternsPackage.Literals.MAP_ENTRY_LINK) {
			destroyMapEntryLink((MapEntryLink) link, matchScope, destroyedObjects);
		}
	}

	@SuppressWarnings("unchecked")
	private void destroyStoryPatternLink(StoryPatternLink link,
			NotifierVariablesScope<Activity, ActivityNode, ActivityEdge, StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> matchScope, Map<AbstractStoryPatternObject, EObject> destroyedObjects) {
		AbstractStoryPatternObject sourceSPO = link.getSource();
		String sourceName = spoFacade.getName(sourceSPO);
		Variable<EClassifier> sourceVar = matchScope.getVariable(sourceName);
		EObject source = sourceVar != null ? (EObject) sourceVar.getValue() : destroyedObjects.get(sourceSPO);
		if(source == null) {
			return;
		}
		
		AbstractStoryPatternObject targetSPO = link.getTarget();
		String targetName = spoFacade.getName(targetSPO);
		Variable<EClassifier> targetVar = matchScope.getVariable(targetName);
		EObject target = targetVar != null ? (EObject) targetVar.getValue() : destroyedObjects.get(targetSPO);
		if(target == null) {
			return;
		}
		
		EStructuralFeature feature = link.getFeature();
		if(feature.isMany()) {
			((Collection<? extends EObject>) source.eGet(feature)).remove(target);
		}
		else {
			source.eSet(feature, null);
		}
		
		getNotificationEmitter().instanceLinkDestroyed(sourceSPO, source, link, targetSPO, target, matchScope, this);
	}

	@SuppressWarnings("unchecked")
	private void destroyContainmentLink(ContainmentLink link,
			NotifierVariablesScope<Activity, ActivityNode, ActivityEdge, StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> matchScope, Map<AbstractStoryPatternObject, EObject> destroyedObjects) {
		AbstractStoryPatternObject sourceSPO = link.getSource();
		String sourceName = spoFacade.getName(sourceSPO);
		Variable<EClassifier> sourceVar = matchScope.getVariable(sourceName);
		EObject source = sourceVar != null ? (EObject) sourceVar.getValue() : destroyedObjects.get(sourceSPO);
		if(source == null) {
			return;
		}
		
		AbstractStoryPatternObject targetSPO = link.getTarget();
		String targetName = spoFacade.getName(targetSPO);
		Variable<EClassifier> targetVar = matchScope.getVariable(targetName);
		EObject target = targetVar != null ? (EObject) targetVar.getValue() : destroyedObjects.get(targetSPO);
		if(target == null) {
			return;
		}
		
		EStructuralFeature feature = target.eContainingFeature();
		EObject container = target.eContainer();
		if(feature == null) {
			return;
		}
		if(feature.isMany()) {
			((Collection<? extends EObject>) container.eGet(feature)).remove(target);
		}
		else {
			container.eSet(feature, null);
		}
		
		getNotificationEmitter().instanceLinkDestroyed(sourceSPO, source, link, targetSPO, target, matchScope, this);
	}

	@SuppressWarnings("unchecked")
	private void destroyMapEntryLink(MapEntryLink link,
			NotifierVariablesScope<Activity, ActivityNode, ActivityEdge, StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> matchScope, Map<AbstractStoryPatternObject, EObject> destroyedObjects) {
		AbstractStoryPatternObject sourceSPO = link.getSource();
		String sourceName = spoFacade.getName(sourceSPO);
		Variable<EClassifier> sourceVar = matchScope.getVariable(sourceName);
		EObject source = sourceVar != null ? (EObject) sourceVar.getValue() : destroyedObjects.get(sourceSPO);
		if(source == null) {
			return;
		}
		
		AbstractStoryPatternObject targetSPO = link.getTarget();
		String targetName = spoFacade.getName(targetSPO);
		Variable<EClassifier> targetVar = matchScope.getVariable(targetName);
		EObject target = targetVar != null ? (EObject) targetVar.getValue() : destroyedObjects.get(targetSPO);
		if(target == null) {
			return;
		}
		
		EStructuralFeature feature = link.getFeature();
		
		final Object value = ((EMap<? extends Object, ? extends Object>) source.eGet(feature)).removeKey(target);
		
		getNotificationEmitter().instanceLinkDestroyed(sourceSPO, source, link, targetSPO, target, matchScope, this);
		
		if(link.getValueTarget() != null && value != null) {
			getNotificationEmitter().instanceLinkDestroyed(sourceSPO, source, link, link.getValueTarget(), value, matchScope, this);
		}
	}

	private void createObjects(NotifierVariablesScope<Activity, ActivityNode, ActivityEdge, StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> matchScope) throws SDMException {
		for (AbstractStoryPatternObject spo : storyPattern.getStoryPatternObjects())
		{
			if (spoFacade.isCreate(spo) && !spoFacade.isOptional(spo) )
			{
				matchScope.deleteVariable(this.spoFacade.getName(spo));
			}
		}
		
		for(AbstractStoryPatternObject spo:storyPattern.getStoryPatternObjects()) {
			if(spo.getModifier() == ModifierEnum.CREATE) {
				createObject(spo, matchScope);
			}
		}
	}

	private void createObject(AbstractStoryPatternObject spo,
			NotifierVariablesScope<Activity, ActivityNode, ActivityEdge, StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> matchScope)
	throws SDMException {
		if (!matchScope.variableExists(spo.getName()))
		{
			final StoryPatternObject storyPatternObject = (StoryPatternObject) spo;
			EClass type = (EClass) spo.getType();

			final EObject eObject;
			if (storyPatternObject.getAssignmentExpression() != null ) {
				
				try {
					eObject = (EObject) expressionInterpreterManager.evaluateExpression(storyPatternObject.getAssignmentExpression(), null, null, matchScope).getValue();
				} catch ( final ExpressionsInterpreterException ex ) {
					throw new SDMException( ex );
				}
			}
			else {
				final EFactory factory = storyPatternObject.getType().getEPackage().getEFactoryInstance();

				eObject = factory.create((EClass) storyPatternObject.getType());
			}
			
			matchScope.createVariable(spo.getName(), type, eObject);
			getNotificationEmitter().instanceObjectCreated(spo, eObject, matchScope, this);
		}
	}

	private void createLinks(NotifierVariablesScope<Activity, ActivityNode, ActivityEdge, StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> matchScope) {
		for(AbstractStoryPatternLink link:storyPattern.getStoryPatternLinks()) {
			if(link.getModifier() == ModifierEnum.CREATE) {
				createLink(link, matchScope);
			}
		}
	}

	private void createLink(AbstractStoryPatternLink link,
			NotifierVariablesScope<Activity, ActivityNode, ActivityEdge, StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> matchScope) {
		if(link.eClass() == MlstorypatternsPackage.Literals.STORY_PATTERN_LINK) {
			createStoryPatternLink((StoryPatternLink) link, matchScope);
		}
		else if(link.eClass() == MlstorypatternsPackage.Literals.CONTAINMENT_LINK) {
			createContainmentLink((ContainmentLink) link, matchScope);
		}
		else if(link.eClass() == MlstorypatternsPackage.Literals.MAP_ENTRY_LINK) {
			createMapEntryLink((MapEntryLink) link, matchScope);
		}
	}

	@SuppressWarnings("unchecked")
	private void createStoryPatternLink(StoryPatternLink link,
			NotifierVariablesScope<Activity, ActivityNode, ActivityEdge, StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> matchScope) {
		//TODO account for all the fancy stuff from MLSDMStoryPatternLinkPatternPart
		AbstractStoryPatternObject sourceSPO = link.getSource();
		String sourceName = spoFacade.getName(sourceSPO);
		Variable<EClassifier> sourceVar = matchScope.getVariable(sourceName);
		if(sourceVar == null) {
			return;
		}
		EObject source = (EObject) sourceVar.getValue();
		
		AbstractStoryPatternObject targetSPO = link.getTarget();
		String targetName = spoFacade.getName(targetSPO);
		Variable<EClassifier> targetVar = matchScope.getVariable(targetName);
		if(targetVar == null) {
			return;
		}
		Object target = targetVar.getValue();
		
		EStructuralFeature feature = link.getFeature();
		if(feature.isMany() && !((Collection<Object>) source.eGet(feature)).contains(target)) {
			((Collection<Object>) source.eGet(feature)).add(target);
		}
		else if(!feature.isMany() && source.eGet(feature) != target){
			source.eSet(feature, target);
		}

		getNotificationEmitter().instanceLinkCreated(sourceSPO, source, link, targetSPO, target, matchScope, this);
	}

	private void createContainmentLink(ContainmentLink link,
			NotifierVariablesScope<Activity, ActivityNode, ActivityEdge, StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> matchScope) {
		//containment links cannot be created
	}

	@SuppressWarnings("unchecked")
	private void createMapEntryLink(MapEntryLink link,
			NotifierVariablesScope<Activity, ActivityNode, ActivityEdge, StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> matchScope) {
		AbstractStoryPatternObject sourceSPO = link.getSource();
		String sourceName = spoFacade.getName(sourceSPO);
		Variable<EClassifier> sourceVar = matchScope.getVariable(sourceName);
		if(sourceVar == null) {
			return;
		}
		EObject source = (EObject) sourceVar.getValue();
		
		AbstractStoryPatternObject targetSPO = link.getTarget();
		String targetName = spoFacade.getName(targetSPO);
		Variable<EClassifier> targetVar = matchScope.getVariable(targetName);
		if(targetVar == null) {
			return;
		}
		Object target = targetVar.getValue();
		
		EStructuralFeature feature = link.getFeature();

		Object valueTarget = null;
		AbstractStoryPatternObject valueTargetSPO = link.getValueTarget();
		if(valueTargetSPO != null) {
			String valueTargetName = spoFacade.getName(valueTargetSPO);
			Variable<EClassifier> valueTargetVar = matchScope.getVariable(valueTargetName);
			if(valueTargetVar != null) {
				valueTarget = valueTargetVar.getValue();
			}
		}
		
		((EMap<Object, Object>) source.eGet(feature)).put(target, valueTarget);

		getNotificationEmitter().instanceLinkCreated(sourceSPO, source, link, targetSPO, target, matchScope, this);
		
		if(valueTarget != null) {
			getNotificationEmitter().instanceLinkCreated(sourceSPO, source, link, valueTargetSPO, valueTarget, matchScope, this);
		}
	}

	private void assignAttributeValues(
			Map<AttributeAssignment, Object> assignmentValues,
			NotifierVariablesScope<Activity, ActivityNode, ActivityEdge, StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> matchScope) {
		for(Entry<AttributeAssignment, Object> entry:assignmentValues.entrySet()) {
			AbstractStoryPatternObject spo = entry.getKey().getStoryPatternObject();
			String name = spoFacade.getName(spo);
			Variable<EClassifier> var = matchScope.getVariable(name);
			if(var == null) {
				continue;
			}
			
			EObject instance = (EObject) var.getValue();
			instance.eSet(entry.getKey().getFeature(), entry.getValue());
			
			this.getNotificationEmitter().attributeValueSet(spo, instance, entry.getKey().getFeature(), entry.getValue(), matchScope, this);
		}
	}

}
