package de.mdelab.mlsdm.interpreter.eclipse;

import org.eclipse.emf.ecore.EClassifier;
import org.eclipse.emf.ecore.EStructuralFeature;

import de.mdelab.expressions.interpreter.core.ExpressionsInterpreterException;
import de.mdelab.mlexpressions.MLExpression;
import de.mdelab.mlsdm.Activity;
import de.mdelab.mlsdm.ActivityEdge;
import de.mdelab.mlsdm.ActivityNode;
import de.mdelab.mlsdm.interpreter.facade.MLSDMMetamodelFacadeFactory;
import de.mdelab.mlsdm.interpreter.notifications.MLSDMNotificationEmitter;
import de.mdelab.mlsdm.interpreter.patternMatcher.patternPartBased.MLSDMPatternPartBasedMatcher;
import de.mdelab.mlstorypatterns.AbstractStoryPatternLink;
import de.mdelab.mlstorypatterns.AbstractStoryPatternObject;
import de.mdelab.mlstorypatterns.StoryPattern;
import de.mdelab.sdm.interpreter.core.SDMException;
import de.mdelab.sdm.interpreter.core.eclipse.MLSDMEclipseExpressionsInterpreterManager;
import de.mdelab.sdm.interpreter.core.facade.MetamodelFacadeFactory;
import de.mdelab.sdm.interpreter.core.variables.NotifierVariablesScope;

/**
 * Implementation of the pattern part based matcher for the StoryDiagramEcore
 * metamodel and use in Eclipse.
 * 
 * @author stephan
 * 
 */
public class MLSDMEclipsePatternPartBasedMatcher extends MLSDMPatternPartBasedMatcher {
	
	public MLSDMEclipsePatternPartBasedMatcher(
			final StoryPattern storyPattern,
			final NotifierVariablesScope<Activity, ActivityNode, ActivityEdge, StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> variableScope,
			final ClassLoader classLoader )
	throws ExpressionsInterpreterException, SDMException	{
		this(	storyPattern,
				variableScope,
				classLoader,
				MLSDMMetamodelFacadeFactory.INSTANCE,
				new MLSDMNotificationEmitter() );
	}
	public MLSDMEclipsePatternPartBasedMatcher(
			final StoryPattern storyPattern,
			final MetamodelFacadeFactory<Activity, ActivityNode, ActivityEdge, StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> p_facadeFactory,
			final NotifierVariablesScope<Activity, ActivityNode, ActivityEdge, StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> variableScope,
			final ClassLoader classLoader) 
	throws ExpressionsInterpreterException, SDMException	{
		this(	storyPattern,
				variableScope,
				classLoader,
				p_facadeFactory,
				new MLSDMNotificationEmitter() );
	}

	public MLSDMEclipsePatternPartBasedMatcher(
			final StoryPattern storyPattern,
			final NotifierVariablesScope<Activity, ActivityNode, ActivityEdge, StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> variablesScope,
			final ClassLoader classLoader,
			final MetamodelFacadeFactory<Activity, ActivityNode, ActivityEdge, StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> facadeFactory,
			final MLSDMNotificationEmitter notificationEmitter )
	throws ExpressionsInterpreterException, SDMException {
		super(	storyPattern,
				variablesScope,
				facadeFactory,
				new MLSDMEclipseExpressionsInterpreterManager<Activity, ActivityNode, ActivityEdge, StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>( MLSDMMetamodelFacadeFactory.INSTANCE.getExpressionFacade(), classLoader, notificationEmitter ).registerExpressionInterpreters(),
				notificationEmitter);
	}
}
