package de.mdelab.mlsdm.interpreter.eclipse.searchModel;

import org.eclipse.emf.ecore.EClassifier;
import org.eclipse.emf.ecore.EStructuralFeature;

import de.mdelab.expressions.interpreter.core.ExpressionsInterpreterException;
import de.mdelab.mlexpressions.MLExpression;
import de.mdelab.mlsdm.Activity;
import de.mdelab.mlsdm.ActivityEdge;
import de.mdelab.mlsdm.ActivityNode;
import de.mdelab.mlsdm.interpreter.facade.MLSDMMetamodelFacadeFactory;
import de.mdelab.mlsdm.interpreter.notifications.MLSDMNotificationEmitter;
import de.mdelab.mlsdm.interpreter.searchModel.MLSDMSearchModelBasedInterpreter;
import de.mdelab.mlstorypatterns.AbstractStoryPatternLink;
import de.mdelab.mlstorypatterns.AbstractStoryPatternObject;
import de.mdelab.mlstorypatterns.StoryPattern;
import de.mdelab.sdm.interpreter.core.SDMException;
import de.mdelab.sdm.interpreter.core.eclipse.MLSDMEclipseExpressionsInterpreterManager;
import de.mdelab.sdm.interpreter.core.facade.MetamodelFacadeFactory;

/**
 * This is an implementation of the search-model-based story diagram
 * interpreter for the MLSDM meta model and use in Eclipse.
 * 
 * @author Matthias.Barkowsky
 *
 */
public class MLSDMEclipseSearchModelBasedInterpreter extends MLSDMSearchModelBasedInterpreter {

	public MLSDMEclipseSearchModelBasedInterpreter(	final ClassLoader classLoader,
													final MLSDMNotificationEmitter notificationEmitter,
													final MetamodelFacadeFactory<Activity, ActivityNode, ActivityEdge, StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> p_facadeFactory )
	throws ExpressionsInterpreterException, SDMException {
		super( new MLSDMEclipseExpressionsInterpreterManager<Activity, ActivityNode, ActivityEdge, StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>( MLSDMMetamodelFacadeFactory.INSTANCE.getExpressionFacade(), classLoader, notificationEmitter).registerExpressionInterpreters() );
	}

	public MLSDMEclipseSearchModelBasedInterpreter(final ClassLoader classLoader)
	throws ExpressionsInterpreterException, SDMException {
		this( classLoader, new MLSDMNotificationEmitter(), MLSDMMetamodelFacadeFactory.INSTANCE );
	}

	public MLSDMEclipseSearchModelBasedInterpreter(	final ClassLoader classLoader,
									final MetamodelFacadeFactory<Activity, ActivityNode, ActivityEdge, StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> p_facadeFactory )
	throws ExpressionsInterpreterException, SDMException {
		this( classLoader, new MLSDMNotificationEmitter(), p_facadeFactory );
	}
}
