package de.mdelab.mlsdm.interpreter.facade;

import de.mdelab.mlsdm.ActivityEdge;
import de.mdelab.mlsdm.ActivityNode;
import de.mdelab.mlsdm.StoryNode;
import de.mdelab.mlsdm.StoryNodeSemanticsEnum;
import de.mdelab.mlstorypatterns.StoryPattern;
import de.mdelab.sdm.interpreter.core.facade.EForEachSemantics;
import de.mdelab.sdm.interpreter.core.facade.IStoryNodeFacade;

/**
 * 
 * @author Stephan Hildebrandt
 * 
 */
public class MLSDMStoryNodeFacade extends MLSDMActivityNodeFacade implements IStoryNodeFacade<ActivityNode, ActivityEdge, StoryPattern>
{

	@Override
	public StoryPattern getStoryPattern(final ActivityNode activityNode)
	{
		assert activityNode != null;
		assert activityNode instanceof StoryNode;

		return ((StoryNode) activityNode).getStoryPattern();
	}

	@Override
	public EForEachSemantics getForEachSemantics(final ActivityNode activityNode)
	{
		assert activityNode != null;
		assert activityNode instanceof StoryNode;

		final StoryNode storyNode = (StoryNode) activityNode;

		switch (storyNode.getSemantics())
		{
			case SINGLE_MATCH:
				return EForEachSemantics.SINGLE_MATCH;
			case FOR_EACH_FRESH_MATCH:
				return EForEachSemantics.FRESH_MATCH;
			
			// Issue #2092
			case FOR_EACH_PRE_SELECT:
				return EForEachSemantics.PRE_SELECT;
			default:
				throw new UnsupportedOperationException();
		}
	}

	@Override
	public ActivityEdge getSuccessNextEdge(final ActivityNode activityNode)
	{
		assert activityNode != null;
		assert activityNode instanceof StoryNode;
		assert activityNode.getOutgoingEdges().size() >= 1;
		assert activityNode.getOutgoingEdges().size() <= 2;

		if (activityNode.getOutgoingEdges().size() == 1 && ((StoryNode) activityNode).getSemantics() == StoryNodeSemanticsEnum.SINGLE_MATCH)
		{
			return activityNode.getOutgoingEdges().get(0);
		}
		else
		{
			for (final ActivityEdge edge : activityNode.getOutgoingEdges())
			{
				switch (edge.getGuardType())
				{
					case SUCCESS:
					case FOR_EACH:
						return edge;

					case FAILURE:
					case END:
						break;

					default:
						throw new UnsupportedOperationException();
				}
			}

			return null;
		}
	}

	@Override
	public ActivityEdge getFailureNextEdge(final ActivityNode activityNode)
	{
		assert activityNode != null;
		assert activityNode instanceof StoryNode;
		assert activityNode.getOutgoingEdges().size() >= 1;
		assert activityNode.getOutgoingEdges().size() <= 2;

		if (activityNode.getOutgoingEdges().size() == 1 && ((StoryNode) activityNode).getSemantics() == StoryNodeSemanticsEnum.SINGLE_MATCH)
		{
			return activityNode.getOutgoingEdges().get(0);
		}
		else
		{
			for (final ActivityEdge edge : activityNode.getOutgoingEdges())
			{
				switch (edge.getGuardType())
				{
					case FAILURE:
					case END:
						return edge;

					case SUCCESS:
					case FOR_EACH:
						break;

					default:
						throw new UnsupportedOperationException();
				}
			}

			return null;
		}

	}
}
