package de.mdelab.mlsdm.interpreter.facade;

import java.util.Collection;
import java.util.HashMap;
import java.util.Map;

import org.eclipse.emf.ecore.EClassifier;
import org.eclipse.emf.ecore.EStructuralFeature;

import de.mdelab.mlexpressions.MLExpression;
import de.mdelab.mlstorypatterns.AbstractStoryPatternLink;
import de.mdelab.mlstorypatterns.AbstractStoryPatternObject;
import de.mdelab.mlstorypatterns.AttributeAssignment;
import de.mdelab.mlstorypatterns.BindingTypeEnum;
import de.mdelab.mlstorypatterns.MatchTypeEnum;
import de.mdelab.mlstorypatterns.ModifierEnum;
import de.mdelab.mlstorypatterns.StoryPatternObject;
import de.mdelab.sdm.interpreter.core.facade.IStoryPatternObjectFacade;

/**
 * 
 * @author Stephan Hildebrandt
 * 
 */
public class MLSDMStoryPatternObjectFacade implements
		IStoryPatternObjectFacade<AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>
{
	@Override
	public String getName(final AbstractStoryPatternObject spo)
	{
		assert spo != null;

		return spo.getName();
	}

	@Override
	public EClassifier getClassifier(final AbstractStoryPatternObject spo)
	{
		assert spo != null;

		return spo.getType();
	}

	@Override
	public boolean isBound(final AbstractStoryPatternObject spo)
	{
		assert spo != null;
		assert spo instanceof StoryPatternObject;

		return ((StoryPatternObject) spo).getBindingType() == BindingTypeEnum.BOUND;
	}

	@Override
	public boolean isMaybeBound(final AbstractStoryPatternObject spo)
	{
		assert spo != null;
		assert spo instanceof StoryPatternObject;

		return ((StoryPatternObject) spo).getBindingType() == BindingTypeEnum.CAN_BE_BOUND;
	}

	@Override
	public boolean isCreate(final AbstractStoryPatternObject spo)
	{
		assert spo != null;

		return spo.getModifier() == ModifierEnum.CREATE;
	}

	@Override
	public boolean isDestroy(final AbstractStoryPatternObject spo)
	{
		return spo.getModifier() == ModifierEnum.DESTROY;
	}

	@Override
	public boolean isOptional(final AbstractStoryPatternObject spo)
	{
		return spo.getMatchType() == MatchTypeEnum.OPTIONAL;
	}

	@Override
	public Map<EStructuralFeature, MLExpression> getAttributeAssignments(final AbstractStoryPatternObject spo)
	{
		assert spo != null;
		assert spo instanceof StoryPatternObject;

		final StoryPatternObject s = (StoryPatternObject) spo;

		if (!s.getAttributeAssignments().isEmpty())
		{
			final Map<EStructuralFeature, MLExpression> map = new HashMap<EStructuralFeature, MLExpression>();

			for (final AttributeAssignment assignment : ((StoryPatternObject) spo).getAttributeAssignments())
			{
				map.put(assignment.getFeature(), assignment.getAssignmentExpression());
			}

			return map;
		}

		return null;
	}

	@Override
	public MLExpression getAssignmentExpression(final AbstractStoryPatternObject spo)
	{
		assert spo != null;
		assert spo instanceof StoryPatternObject;

		return ((StoryPatternObject) spo).getAssignmentExpression();
	}

	@Override
	public Collection<MLExpression> getConstraints(final AbstractStoryPatternObject spo)
	{
		assert spo != null;
		assert spo instanceof StoryPatternObject;

		return ((StoryPatternObject) spo).getConstraints();
	}
}
