package de.mdelab.mlsdm.interpreter.searchModel.model.action;

import java.util.Collections;
import java.util.Iterator;
import java.util.Map;

import org.eclipse.emf.ecore.EClassifier;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EStructuralFeature;

import de.mdelab.expressions.interpreter.core.variables.Variable;
import de.mdelab.mlexpressions.MLExpression;
import de.mdelab.mlsdm.interpreter.searchModel.model.MLSDMMatchingAction;
import de.mdelab.mlsdm.interpreter.searchModel.patternMatcher.MLSDMSearchModelBasedMatcher;
import de.mdelab.mlstorypatterns.AbstractStoryPatternLink;
import de.mdelab.mlstorypatterns.AbstractStoryPatternObject;
import de.mdelab.mlstorypatterns.StoryPattern;
import de.mdelab.mlstorypatterns.StoryPatternLink;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.MatchSingleNodeMatchingTransaction;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.MatchState;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.MatchingAction;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.PatternConstraint;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.PatternNode;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.SearchModelMatchingTransaction;

public class MLSDMLinkContainmentReverseMatchingAction extends MLSDMMatchingAction {

	/*
	 * The link.
	 */
	private StoryPatternLink link;
	
	/*
	 * The link type.
	 */
	private EStructuralFeature feature;
	
	/*
	 * Source node of the link.
	 */
	private PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> sourceNode;
	
	/*
	 * Target node of the link.
	 */
	private PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> targetNode;
	
	
	private MatchSingleNodeMatchingTransaction<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> transaction;
	
	/**
	 * This is a match state specific to the traverse link
	 * matching action.
	 * 
	 * @author Matthias.Barkowsky
	 *
	 */
	private static class MLSDMLinkContainmentReverseMatchState extends MatchState {

		/*
		 * Iterator of possible link targets in the instance model.
		 */
		private Iterator<? extends Object> featureIterator;
		
		private MLSDMLinkContainmentReverseMatchState(Iterator<? extends Object> featureIterator) {
			this.featureIterator = featureIterator;
		}

		private Iterator<? extends Object> getFeatureIterator() {
			return featureIterator;
		}
		
	}
	
	public MLSDMLinkContainmentReverseMatchingAction(int id,
			MLSDMSearchModelBasedMatcher matcher,
			PatternConstraint<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> patternConstraint,
			StoryPatternLink link,
			EStructuralFeature feature,
			PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> sourceNode,
			PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> targetNode) {
		super(id, matcher, patternConstraint);
		this.link = link;
		this.feature = feature;
		this.sourceNode = sourceNode;
		this.targetNode = targetNode;
		this.addRequiredBinding(targetNode);
		initializeTransaction();
	}

	/**
	 * Create and initialize a matching transaction for this matching action.
	 * The created transaction can be reused to avoid overhead since for each
	 * matching action, only one transaction can be on the matching stack at
	 * a time.
	 */
	private void initializeTransaction() {
		transaction = new MatchSingleNodeMatchingTransaction<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>(patternConstraint, matcher);

		transaction.setSPO(sourceNode.getSpo());
		transaction.addPatternNode(sourceNode);
	}

	/* (non-Javadoc)
	 * @see de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.MatchingAction#doEstimateCardinality()
	 */
	@Override
	protected double doEstimateCardinality()  {
		if(!sourceNode.isBound() && targetNode.isBound()) {
			AbstractStoryPatternObject targetSPO = targetNode.getSpo();
		
			Variable<EClassifier> targetVariable = matcher.getVariablesScope().getVariable(targetSPO.getName());
			
			return ((EObject) targetVariable.getValue()).eContainer() != null ? 1 : 0;
		}
		else {
			return PatternConstraint.MATCHING_NOT_POSSIBLE;
		}
	}

	/* (non-Javadoc)
	 * @see de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.MatchingAction#createMatchState()
	 */
	@Override
	public MatchState createMatchState() {
		AbstractStoryPatternObject targetSPO = targetNode.getSpo();
		Variable<EClassifier> targetVariable = matcher.getVariablesScope().getVariable(targetSPO.getName());
		EObject eContainer = ((EObject) targetVariable.getValue()).eContainer();
		
		Iterator<? extends Object> iterator;
		if(eContainer != null) {
			iterator = Collections.singleton(eContainer).iterator();
		}
		else {
			iterator = Collections.emptyList().iterator();
		}
		
		return new MLSDMLinkContainmentReverseMatchState(iterator);
	}

	/* (non-Javadoc)
	 * @see de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.MatchingAction#match(de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.MatchState)
	 */
	@Override
	public SearchModelMatchingTransaction<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> match(
			MatchState matchState) {
		EObject instanceObject = (EObject) matcher.getVariable(targetNode.getSpo().getName()).getValue();
		matcher.getNotificationEmitter().traversingLink(link, link.getSource(), instanceObject, link.getTarget(), matcher.getVariablesScope(), matcher);
		
		boolean valid = false;
		MLSDMLinkContainmentReverseMatchState ms = (MLSDMLinkContainmentReverseMatchState) matchState;
		Iterator<? extends Object> featureIterator = ms.getFeatureIterator();
		Object sourceInstance = null;
		while(!valid && featureIterator.hasNext()) {
			sourceInstance = featureIterator.next();
			valid = matcher.instanceMatchesSPO(sourceInstance, sourceNode.getSpo()) && matcher.checkMatchingHistory(this.getPatternConstraint(), sourceInstance);
		}
		
		transaction.setValid(valid);
		
		if(valid) {
			transaction.setTargetInstance(sourceInstance);
		}
		else {
			matcher.getNotificationEmitter().storyPatternObjectNotBound(sourceNode.getSpo(), matcher.getVariablesScope(),
					matcher);
		}
		

		matcher.updateMatchingHistory(this.getPatternConstraint(), sourceInstance);
		
		return transaction;
	}

	public EStructuralFeature getFeature() {
		return feature;
	}
	
	public PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> getSourceNode() {
		return sourceNode;
	}

	public PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> getTargetNode() {
		return targetNode;
	}
	
	/* (non-Javadoc)
	 * @see java.lang.Object#toString()
	 */
	@Override
	public String toString() {
		return "traverse\t(" + targetNode.getSpo().getName() + " <-" + feature.getName() + "- " + sourceNode.getSpo().getName() + ")";
	}

	@SuppressWarnings("unchecked")
	@Override
	public MatchingAction<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> clonePristine(
			PatternConstraint<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> clonePatternConstraint,
			Map<Object, Object> clonePNs) {
		return new MLSDMLinkContainmentReverseMatchingAction(id, matcher, clonePatternConstraint, link, feature, (PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>) clonePNs.get(sourceNode), (PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>) clonePNs.get(targetNode));
	}

}
