package de.mdelab.mlsdm.interpreter.searchModel.model.action;

import java.util.Iterator;
import java.util.Map;
import java.util.Map.Entry;

import org.eclipse.emf.common.util.EMap;
import org.eclipse.emf.ecore.EClassifier;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EStructuralFeature;

import de.mdelab.mlexpressions.MLExpression;
import de.mdelab.mlsdm.interpreter.searchModel.model.MLSDMMatchingAction;
import de.mdelab.mlsdm.interpreter.searchModel.patternMatcher.MLSDMSearchModelBasedMatcher;
import de.mdelab.mlstorypatterns.AbstractStoryPatternLink;
import de.mdelab.mlstorypatterns.AbstractStoryPatternObject;
import de.mdelab.mlstorypatterns.MapEntryLink;
import de.mdelab.mlstorypatterns.StoryPattern;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.MatchMultipleNodesMatchingTransaction;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.MatchSingleNodeMatchingTransaction;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.MatchState;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.MatchingAction;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.PatternConstraint;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.PatternNode;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.SearchModelMatchingTransaction;

/**
 * This is a matching action for matching both key and value
 * node of a map entry link in the story pattern. It can be
 * executed once a mapping for the source node of the link
 * has been created.
 * 
 * @author Matthias.Barkowsky
 *
 */
public class MLSDMMapEntryMatchingAction extends MLSDMMatchingAction {

	/**
	 * This is a match state specific to the map entry matching
	 * action.
	 * 
	 * @author Matthias.Barkowsky
	 *
	 */
	private class MLSDMMapEntryMatchState extends MatchState{

		/*
		 * Iterator of all candidate map entries.
		 */
		private Iterator<Entry<Object, Object>> iterator;

		private MLSDMMapEntryMatchState(Iterator<Entry<Object, Object>> iterator) {
			this.iterator = iterator;
		}

		private Iterator<Entry<Object, Object>> getIterator() {
			return iterator;
		}
		
	}
	
	/*
	 * The map entry link.
	 */
	private MapEntryLink link;
	
	/*
	 * Source node of the link.
	 */
	private PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> sourceNode;
	
	/*
	 * Value node of the link.
	 */
	private PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> valueNode;
	
	/*
	 * Key node of the link.
	 */
	private PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> keyNode;
	
	public MLSDMMapEntryMatchingAction(int id,
			MLSDMSearchModelBasedMatcher matcher, MapEntryLink link,
			PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> sourceNode,
			PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> valueNode,
			PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> keyNode ) {
		super(id, matcher);
		this.link = link;
		this.sourceNode = sourceNode;
		this.valueNode = valueNode;
		this.keyNode = keyNode;
		addRequiredBinding(sourceNode);
	}
	
	/* (non-Javadoc)
	 * @see de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.MatchingAction#createMatchState()
	 */
	@SuppressWarnings("unchecked")
	@Override
	public MatchState createMatchState() {
		if(!sourceNode.isBound()) {
			return null;
		}
		else {
			EObject sourceInstance = (EObject) matcher.getVariable(sourceNode.getSpo().getName()).getValue();
			return new MLSDMMapEntryMatchState(((EMap<Object, Object>) sourceInstance.eGet(link.getFeature())).iterator());
		}
	}

	/* (non-Javadoc)
	 * @see de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.MatchingAction#doEstimateCardinality()
	 */
	@SuppressWarnings("unchecked")
	@Override
	protected double doEstimateCardinality() {
		if(!sourceNode.isBound()) {
			return PatternConstraint.MATCHING_NOT_POSSIBLE;
		}
		else {
			EObject sourceInstance = (EObject) matcher.getVariable(sourceNode.getSpo().getName()).getValue();
			return ((EMap<Object, Object>) sourceInstance.eGet(link.getFeature())).size();
		}
	}

	/* (non-Javadoc)
	 * @see de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.MatchingAction#match(de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.MatchState)
	 */
	@Override
	public SearchModelMatchingTransaction<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> match(
			MatchState matchState) {
		MatchMultipleNodesMatchingTransaction<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> transaction =
				new MatchMultipleNodesMatchingTransaction<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>(getPatternConstraint());
		
		EObject instanceObject = (EObject) matcher.getVariable(sourceNode.getSpo().getName()).getValue();
		matcher.getNotificationEmitter().traversingLink(link, link.getSource(), instanceObject, link.getTarget(), matcher.getVariablesScope(), matcher);
		
		if(!(matchState instanceof MLSDMMapEntryMatchState)) {
			transaction.setValid(false);
		}
		else {
			boolean valid = false;
			MLSDMMapEntryMatchState ms = (MLSDMMapEntryMatchState) matchState;
			Iterator<Entry<Object, Object>> featureIterator = ms.getIterator();
			Entry<Object, Object> entry = null;
			
			while(!valid && featureIterator.hasNext()) {
				transaction = new MatchMultipleNodesMatchingTransaction<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>(getPatternConstraint());
				
				entry = featureIterator.next();

				if(!matcher.checkMatchingHistory(this.getPatternConstraint(), entry)) {
					valid = false;
					transaction.setValid(valid);
					continue;
				}
				
				Object value = entry.getValue();
				Object key = entry.getKey();
				
				//ensure isomorphism
				if(value == key) {
					valid = false;
					transaction.setValid(valid);
					continue;
				}

				boolean valueBound = valueNode != null ? valueNode.isBound() : false;
				boolean keyBound = keyNode.isBound();

				//check if value either matches already bound instance or is a valid new binding
				if(valueNode == null) {
					valid = true;
				}
				else if(valueBound && matcher.getVariable(valueNode.getSpo().getName()).getValue() == value) {
					 valid = true;
				}
				else if(!valueBound && matcher.instanceMatchesSPO(value, valueNode.getSpo())) {
					valid = true;
					MatchSingleNodeMatchingTransaction<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> matchValueTransaction =
							new MatchSingleNodeMatchingTransaction<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>(getPatternConstraint(), matcher);
					matchValueTransaction.setValid(valid);
					matchValueTransaction.setSPO(valueNode.getSpo());
					matchValueTransaction.setTargetInstance(value);
					
					transaction.addSingleNodeTransaction(matchValueTransaction);
					transaction.addPatternNode(valueNode);
				}
				else {
					valid = false;
				}
				
				//check if key either matches already bound instance or is a valid new binding
				if(valid && keyBound && matcher.getVariable(keyNode.getSpo().getName()).getValue() == key) {
					 valid = true;
				}
				else if(valid && !keyBound && matcher.instanceMatchesSPO(key, keyNode.getSpo())) {
					valid = true;
					MatchSingleNodeMatchingTransaction<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> matchKeyTransaction =
							new MatchSingleNodeMatchingTransaction<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>(getPatternConstraint(), matcher);
					matchKeyTransaction.setValid(valid);
					matchKeyTransaction.setSPO(keyNode.getSpo());
					matchKeyTransaction.setTargetInstance(key);
					
					transaction.addSingleNodeTransaction(matchKeyTransaction);
					transaction.addPatternNode(keyNode);
				}
				else {
					valid = false;
				}				
			}
			
			transaction.setValid(valid);

			matcher.updateMatchingHistory(getPatternConstraint(), entry);
		}
		return transaction;
	}

	/* (non-Javadoc)
	 * @see java.lang.Object#toString()
	 */
	@Override
	public String toString() {
		return "traverse \t\t (" + link.getSource().getName() + " -" + link.getFeature().getName() + "-> " + "(" + link.getTarget().getName() + ", " + (link.getValueTarget() != null ? link.getValueTarget().getName() : null) + "))";
	}

	@SuppressWarnings("unchecked")
	@Override
	public MatchingAction<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> clonePristine(
			PatternConstraint<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> clonePatternConstraint,
			Map<Object, Object> clonePNs) {
		return new MLSDMMapEntryMatchingAction(id, matcher, link, (PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>) clonePNs.get(sourceNode),(PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>)  clonePNs.get(valueNode),(PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>)  clonePNs.get(keyNode));
	}
}
