package de.mdelab.mlsdm.interpreter.searchModel.patternMatcher;

import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.EClassifier;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EReference;
import org.eclipse.emf.ecore.EStructuralFeature;
import org.eclipse.emf.ecore.EcoreFactory;
import org.eclipse.emf.ecore.EcorePackage;

import de.mdelab.expressions.interpreter.core.ExpressionsInterpreterException;
import de.mdelab.expressions.interpreter.core.variables.Variable;
import de.mdelab.mlexpressions.MLExpression;
import de.mdelab.mlsdm.Activity;
import de.mdelab.mlsdm.ActivityEdge;
import de.mdelab.mlsdm.ActivityNode;
import de.mdelab.mlsdm.MlsdmPackage;
import de.mdelab.mlsdm.StoryNode;
import de.mdelab.mlsdm.interpreter.MLSDMExpressionInterpreterManager;
import de.mdelab.mlsdm.interpreter.searchModel.patternMatcher.expressions.MLSDMAdapterBackedOCLExpressionInterpreter;
import de.mdelab.mlsdm.interpreter.searchModel.patternMatcher.expressions.MLSDMExpressionAnalyzer;
import de.mdelab.mlsdm.interpreter.searchModel.patternMatcher.expressions.MLSDMOCLExpressionAnalyzer;
import de.mdelab.mlstorypatterns.AbstractStoryPatternLink;
import de.mdelab.mlstorypatterns.AbstractStoryPatternObject;
import de.mdelab.mlstorypatterns.StoryPattern;
import de.mdelab.mlstorypatterns.StoryPatternObject;
import de.mdelab.sdm.interpreter.core.notifications.NotificationEmitter;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.expressions.FeatureAccess;
import de.mdelab.sdm.interpreter.core.variables.NotifierVariablesScope;

public class MLSDMMetadataIndex extends MLSDMReferenceIndex {

	public static final int EXPRESSION_NOT_REGISTERED = -1;
	private Map<EReference, Integer> maxCardinalities;
	private Map<EReference, Integer> maxInverseCardinalities;
	private Map<EReference, Integer> minCardinalities;
	private Map<EReference, Integer> minInverseCardinalities;
	private Map<MLExpression, Integer> validInstanceNumbers;		//TODO this needs to be maintained somehow
	private MLSDMExpressionInterpreterManager expressionInterpreterManager;
	
	public MLSDMMetadataIndex() {
		super();
		this.maxCardinalities = new HashMap<EReference, Integer>();
		this.maxInverseCardinalities = new HashMap<EReference, Integer>();
		this.minCardinalities = new HashMap<EReference, Integer>();
		this.minInverseCardinalities = new HashMap<EReference, Integer>();
		this.validInstanceNumbers = new HashMap<MLExpression, Integer>();
		this.expressionInterpreterManager = new MLSDMExpressionInterpreterManager(MLSDMMetadataIndex.class.getClassLoader());
		//try {
		this.expressionInterpreterManager.registerExpressionInterpreter(new MLSDMAdapterBackedOCLExpressionInterpreter(this), "OCL", "1.0");
//		} catch (SDMException e) {
//			e.printStackTrace();
//		}
	}
	
	@Override
	public void addReference(EObject source, EReference reference, EObject target) {
		super.addReference(source, reference, target);
		
		int cardinality;
		if(reference.isMany()) {
			cardinality = ((Collection<?>) source.eGet(reference)).size();
		}
		else {
			cardinality = 1;
		}
		if(!maxCardinalities.containsKey(reference) || cardinality > maxCardinalities.get(reference)) {
			maxCardinalities.put(reference, cardinality);
		}
//		if(!minCardinalities.containsKey(reference) || cardinality < minCardinalities.get(reference)) {	//TODO obviously cannot be maintained this way...
//			minCardinalities.put(reference, cardinality);
//		}

		int inverseCardinality = getInverseReferences(target, reference).size();
		if(!maxInverseCardinalities.containsKey(reference) || inverseCardinality > maxInverseCardinalities.get(reference)) {
			maxInverseCardinalities.put(reference, inverseCardinality);
		}
//		if(!minInverseCardinalities.containsKey(reference) || inverseCardinality < minInverseCardinalities.get(reference)) {	//TODO obviously cannot be maintained this way...
//			minInverseCardinalities.put(reference, inverseCardinality);
//		}
	}
	
	public int getMaxCardinality(EReference reference) {
		return maxCardinalities.containsKey(reference) ? maxCardinalities.get(reference) : 0;
	}

	public int getMinCardinality(EReference reference) {
		return minCardinalities.containsKey(reference) ? minCardinalities.get(reference) : 0;
	}

	public int getMaxInverseCardinality(EReference reference) {
		return maxInverseCardinalities.containsKey(reference) ? maxInverseCardinalities.get(reference) : 0;
	}

	public int getMinInverseCardinality(EReference reference) {
		return minInverseCardinalities.containsKey(reference) ? minInverseCardinalities.get(reference) : 0;
	}
	
	public int getValidInstanceNumbers(Object expression) {
		if(!validInstanceNumbers.containsKey(expression)) {
			return EXPRESSION_NOT_REGISTERED;
		}
		else {
			return validInstanceNumbers.get(expression);
		}
	}
	
	public void registerExpression(MLExpression expression, EClass contextClass,
			NotifierVariablesScope<Activity, ActivityNode, ActivityEdge, StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> vs) {
		Collection<EObject> domain = getDomain(contextClass);
		int validInstanceNumber = 0;
		for(EObject context:domain) {
			try {
				if((Boolean) expressionInterpreterManager.evaluateExpression(expression, contextClass, context, vs).getValue()) {
					validInstanceNumber++;
				}
			} catch (ExpressionsInterpreterException e) {
				e.printStackTrace();
			}
		}
		validInstanceNumbers.put(expression, validInstanceNumber);
	}

	private void registerExpression(MLExpression expression, EClass contextClass, String spoName,
			NotifierVariablesScope<Activity, ActivityNode, ActivityEdge, StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> vs) {
		Collection<EObject> domain = getDomain(contextClass);
		int validInstanceNumber = 0;
		vs = vs.clone();
		for(EObject context:domain) {
			vs.createVariable(spoName, contextClass, context);
			try {
				if((Boolean) expressionInterpreterManager.evaluateExpression(expression, contextClass, null, vs).getValue()) {
					validInstanceNumber++;
				}
			} catch (ExpressionsInterpreterException e) {
				e.printStackTrace();
			}
		}
		validInstanceNumbers.put(expression, validInstanceNumber);
	}

	public void registerExpressions(Activity activity, Collection<Variable<EClassifier>> inputVariables, boolean print) {
		NotifierVariablesScope<Activity, ActivityNode, ActivityEdge, StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> vs
				= new NotifierVariablesScope<Activity, ActivityNode, ActivityEdge, StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>(
						new NotificationEmitter<Activity, ActivityNode, ActivityEdge, StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>());
		for(Variable<EClassifier> variable:inputVariables) {
			vs.createVariable(variable.getName(), variable.getClassifier(), variable.getValue());
		}
		
		for(ActivityNode node:activity.getNodes()) {
			if(node.eClass() == MlsdmPackage.Literals.STORY_NODE) {
				registerExpressions(((StoryNode)node).getStoryPattern(), vs, print);
			}
		}
	}

	public void registerExpressions(StoryPattern pattern,
			Collection<Variable<EClassifier>> variables,
			boolean print) {
		NotifierVariablesScope<Activity, ActivityNode, ActivityEdge, StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> vs
		= new NotifierVariablesScope<Activity, ActivityNode, ActivityEdge, StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>(
				new NotificationEmitter<Activity, ActivityNode, ActivityEdge, StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>());
		for(Variable<EClassifier> variable:variables) {
			vs.createVariable(variable.getName(), variable.getClassifier(), variable.getValue());
		}
		
		registerExpressions(pattern, vs, print);
	}
	
	public void registerExpressions(StoryPattern pattern,
			NotifierVariablesScope<Activity, ActivityNode, ActivityEdge, StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> vs,
			boolean print) {
		for(AbstractStoryPatternObject spo:pattern.getStoryPatternObjects()) {
			if(!(spo.getType().eClass() == EcorePackage.Literals.ECLASS)) {
				continue;
			}
			for(MLExpression expression:((StoryPatternObject) spo).getConstraints()) {
				long start = System.currentTimeMillis();
				registerExpression(expression, (EClass) spo.getType(), vs);
				long end = System.currentTimeMillis();
				if(print) System.out.println(spo.getName() + "_expression_time=" + (end - start));
			}
		}
		
		MLSDMExpressionAnalyzer oclAnalyzer = new MLSDMOCLExpressionAnalyzer();
		NotifierVariablesScope<Activity, ActivityNode, ActivityEdge, StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> dummyScope = addDummyVariables(pattern, vs.clone());
		oclAnalyzer.setVariablesScope(dummyScope);
		
		Map<String, AbstractStoryPatternObject> spoNames = getSPONames(pattern);
		for(MLExpression constraint:pattern.getConstraints()) {
			Collection<FeatureAccess<EStructuralFeature>> accesses = oclAnalyzer.getFeatureAccesses(constraint);
			if(accesses != null) {
				Collection<String> accessedSPONames = getAccessedSpoNames(accesses, spoNames);
				if(accessedSPONames.size() == 1) {
					String spoName = accessedSPONames.iterator().next();
					registerExpression(constraint, (EClass) spoNames.get(spoName).getType(), spoName, vs);
				}
			}
		}
	}

	private NotifierVariablesScope<Activity, ActivityNode, ActivityEdge, StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> addDummyVariables(
			StoryPattern pattern, NotifierVariablesScope<Activity, ActivityNode, ActivityEdge, StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> variablesScope) {
		for(AbstractStoryPatternObject spo:pattern.getStoryPatternObjects()) {
			variablesScope.createVariable(spo.getName(), spo.getType(), EcoreFactory.eINSTANCE.createEObject());
		}
		return variablesScope;
	}

	private Map<String, AbstractStoryPatternObject> getSPONames(StoryPattern pattern) {
		HashMap<String, AbstractStoryPatternObject> spoNames = new HashMap<String, AbstractStoryPatternObject>();
		for(AbstractStoryPatternObject spo:pattern.getStoryPatternObjects()) {
			spoNames.put(spo.getName(), spo);
		}
		return spoNames;
	}

	private Collection<String> getAccessedSpoNames(Collection<FeatureAccess<EStructuralFeature>> accesses, Map<String, AbstractStoryPatternObject> spoNames) {
		Set<String> accessedSPONames = new HashSet<String>();
		for(FeatureAccess<EStructuralFeature> access:accesses) {
			if(spoNames.containsKey(access.getVariableName())) {
				accessedSPONames.add(access.getVariableName());
			}
		}
		return accessedSPONames;
	}

}
