package de.mdelab.mlsdm.interpreter.searchModel.patternMatcher;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Map.Entry;
import java.util.Set;

import org.eclipse.emf.ecore.EClassifier;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EReference;
import org.eclipse.emf.ecore.EStructuralFeature;

import de.mdelab.expressions.interpreter.core.ExpressionInterpreterManager;
import de.mdelab.expressions.interpreter.core.variables.Variable;
import de.mdelab.mlexpressions.MLExpression;
import de.mdelab.mlsdm.Activity;
import de.mdelab.mlsdm.ActivityEdge;
import de.mdelab.mlsdm.ActivityNode;
import de.mdelab.mlsdm.interpreter.MLSDMMatchApplier;
import de.mdelab.mlstorypatterns.AbstractStoryPatternLink;
import de.mdelab.mlstorypatterns.AbstractStoryPatternObject;
import de.mdelab.mlstorypatterns.StoryPattern;
import de.mdelab.sdm.interpreter.core.facade.MetamodelFacadeFactory;
import de.mdelab.sdm.interpreter.core.notifications.NotificationEmitter;
import de.mdelab.sdm.interpreter.core.variables.NotifierVariablesScope;

public class MLSDMSearchModelBasedMatchApplier extends MLSDMMatchApplier {

	protected MLSDMReferenceIndex referenceAdapter;
	
	public MLSDMSearchModelBasedMatchApplier(StoryPattern storyPattern,
			MetamodelFacadeFactory<Activity, ActivityNode, ActivityEdge, StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> facadeFactory,
			ExpressionInterpreterManager<EClassifier, EStructuralFeature, MLExpression> expressionInterpreterManager,
			NotificationEmitter<Activity, ActivityNode, ActivityEdge, StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> notificationEmitter,
			MLSDMReferenceIndex referenceAdapter) {
		super(storyPattern, facadeFactory, expressionInterpreterManager, notificationEmitter);
		this.referenceAdapter = referenceAdapter;
	}

	protected EObject destroyObject(AbstractStoryPatternObject spo, NotifierVariablesScope<Activity, ActivityNode, ActivityEdge, StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> matchScope) {
		EObject destroyedObject = null;
		String name = spoFacade.getName(spo);
		Variable<EClassifier> var = matchScope.getVariable(name);
		if(var != null) {
			matchScope.deleteVariable(name);
			destroyedObject = (EObject) var.getValue();
			deleteObject(destroyedObject);
			
			getNotificationEmitter().instanceObjectDestroyed(spo, destroyedObject, matchScope, this);
		}
		return destroyedObject;
	}

	private void deleteObject(EObject destroyedObject) {
		//use optimized delete if possible
		if(referenceAdapter != null) {
			Collection<Entry<EReference, Set<EObject>>> deletedInverseReferences = new ArrayList<Entry<EReference, Set<EObject>>>(referenceAdapter.getInverseReferences(destroyedObject).entrySet());
			for(Entry<EReference, Set<EObject>> incomingEntry:deletedInverseReferences) {
				for(EObject source:new ArrayList<EObject>(incomingEntry.getValue())) {
					EStructuralFeature feature = incomingEntry.getKey();
					if(!feature.isMany()) {
						source.eSet(feature, null);
					}
					else {
						((Collection<?>) source.eGet(feature)).remove(destroyedObject);
					}
				}
			}

			for(EStructuralFeature feature:destroyedObject.eClass().getEAllStructuralFeatures()) {
				if(!feature.isMany()) {
					destroyedObject.eSet(feature, null);
				}
				else {
					((Collection<?>) destroyedObject.eGet(feature)).clear();
				}
			}
		}
		else {
			instanceFacade.delete(destroyedObject);
		}
		
	}

}
