package de.mdelab.mlsdm.interpreter.searchModel.patternMatcher;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.List;
import java.util.Map;

import org.eclipse.emf.ecore.EClassifier;
import org.eclipse.emf.ecore.EStructuralFeature;

import de.mdelab.mlexpressions.MLExpression;
import de.mdelab.mlexpressions.MLStringExpression;
import de.mdelab.mlsdm.interpreter.searchModel.MLSDMSearchModelBasedInterpreter;
import de.mdelab.mlsdm.interpreter.searchModel.model.MLSDMPatternConstraint;
import de.mdelab.mlsdm.interpreter.searchModel.model.MLSDMPatternNode;
import de.mdelab.mlsdm.interpreter.searchModel.model.action.MLSDMExpressionCheckMatchingAction;
import de.mdelab.mlsdm.interpreter.searchModel.model.action.MLSDMIndexStepMatchingAction;
import de.mdelab.mlstorypatterns.AbstractStoryPatternLink;
import de.mdelab.mlstorypatterns.AbstractStoryPatternObject;
import de.mdelab.mlstorypatterns.StoryPattern;
import de.mdelab.sdm.icl.iCL.ICLConstraint;
import de.mdelab.sdm.icl.iCL.ICLPackage;
import de.mdelab.sdm.icl.iCL.ICLParameter;
import de.mdelab.sdm.icl.iCL.ICLRange;
import de.mdelab.sdm.icl.iCL.ICLValue;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.PatternNode;

/**
 * Search model builder for the MLSDM meta model that splits index matching
 * actions into several individual matching actions for different parameter
 * binding configurations.
 * 
 * @author Matthias.Barkowsky
 *
 */
public class MLSDMStepwiseSearchModelBuilder extends MLSDMSearchModelBuilder{

	/* (non-Javadoc)
	 * @see de.mdelab.mlsdm.interpreter.searchModel.patternMatcher.MLSDMSearchModelBuilder#createStagedAccessIndexConstraints(de.mdelab.sdm.icl.iCL.ICLConstraint, de.mdelab.mlexpressions.MLExpression, java.util.Map, java.util.Map)
	 */
	protected Collection<MLSDMPatternConstraint> createStagedAccessIndexConstraints(ICLConstraint iclConstraint,
			MLExpression expression, Map<AbstractStoryPatternObject, MLSDMPatternNode> patternNodes,
			Map<String, AbstractStoryPatternObject> nonCreateSPOs) {
		List<MLSDMPatternConstraint> patternConstraints = new ArrayList<MLSDMPatternConstraint>();
		
		if(!matcher.getInterpreterParameters().containsKey(MLSDMSearchModelBasedInterpreter.ENABLE_STAGED_INDEX_MATCHING) || (Boolean) matcher.getInterpreterParameters().get(MLSDMSearchModelBasedInterpreter.ENABLE_STAGED_INDEX_MATCHING)) {	
			super.createStagedAccessIndexConstraints(iclConstraint, expression, patternNodes, nonCreateSPOs);
		}
		else {
			/*
			 * Create pattern constraints for regular index-based
			 * matching.
			 */
			MLSDMPatternConstraint patternConstraint = new MLSDMPatternConstraint(iclConstraint);

			Map<ICLValue, List<? extends PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>>> iclValueToPatternNode =
						mapICLValuesToPatternNodes(iclConstraint, nonCreateSPOs, patternNodes);
			
			patternConstraint.addDependencies(flatten(iclValueToPatternNode.values()));
			
			Collection<PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>> requiredBindings =
					new HashSet<PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>>();
			/*
			 * All pattern nodes that are required by a non-variable index parameter
			 * cannot be matched and are hence required by default.
			 */
			for(ICLParameter parameter:iclConstraint.getParameters().getList()) {
				if(!(parameter.eClass() == ICLPackage.Literals.ICL_VARIABLE)) {
					if(parameter.eClass() == ICLPackage.Literals.ICL_RANGE) {
						requiredBindings.addAll(iclValueToPatternNode.get(((ICLRange) parameter).getMin()));
						requiredBindings.addAll(iclValueToPatternNode.get(((ICLRange) parameter).getMax()));
					}
					else {
						requiredBindings.addAll(iclValueToPatternNode.get(parameter));
					}
				}
			}
			
			/*
			 * Create a matching action for each configuration of bindings for variable
			 * parameters (except for explicit check).
			 */
			for(ICLParameter parameter:iclConstraint.getParameters().getList()) {
				/*
				 * Skip non-variable parameters
				 */
				if(!(parameter.eClass() == ICLPackage.Literals.ICL_VARIABLE)) {
					continue;
				}
				
				MLSDMIndexStepMatchingAction matchingAction = new MLSDMIndexStepMatchingAction(currentActionId++, matcher, iclConstraint, (MLStringExpression) expression, iclValueToPatternNode, requiredBindings);
				patternConstraint.addPossibleAction(matchingAction);
				
				if(parameter instanceof ICLRange) {
					requiredBindings.addAll(iclValueToPatternNode.get(((ICLRange) parameter).getMin()));
					requiredBindings.addAll(iclValueToPatternNode.get(((ICLRange) parameter).getMax()));
				}
				else {
					requiredBindings.addAll(iclValueToPatternNode.get(parameter));
				}
			}
			
			MLSDMExpressionCheckMatchingAction explicitCheckAction = new MLSDMExpressionCheckMatchingAction(currentActionId++, matcher, expression, patternConstraint.getDependencies());
			patternConstraint.addPossibleAction(explicitCheckAction);
			patternConstraint.setExplicitCheckAction(explicitCheckAction);
			
			patternConstraints.add(patternConstraint);
		}
		return patternConstraints;
	}

}
