package de.mdelab.mlsdm.interpreter.searchModel.patternMatcher.strategy.grgen;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.EClassifier;
import org.eclipse.emf.ecore.EReference;
import org.eclipse.emf.ecore.EStructuralFeature;

import de.mdelab.mlexpressions.MLExpression;
import de.mdelab.mlsdm.interpreter.searchModel.model.action.MLSDMDomainMatchMatchingAction;
import de.mdelab.mlsdm.interpreter.searchModel.model.action.MLSDMLinkLookupMatchingAction;
import de.mdelab.mlsdm.interpreter.searchModel.model.action.MLSDMLinkTraverseMatchingAction;
import de.mdelab.mlsdm.interpreter.searchModel.model.action.MLSDMLinkTraverseReverseMatchingAction;
import de.mdelab.mlsdm.interpreter.searchModel.patternMatcher.MLSDMMetadataIndex;
import de.mdelab.mlstorypatterns.AbstractStoryPatternLink;
import de.mdelab.mlstorypatterns.AbstractStoryPatternObject;
import de.mdelab.mlstorypatterns.StoryPattern;
import de.mdelab.mlstorypatterns.StoryPatternLink;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.MatchingAction;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.PatternConstraint;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.PatternNode;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.SearchModel;

public class GRGENPlanGraphBuilder {

	private MLSDMMetadataIndex metadataAdapter;
	private GRGENGeometricMeansAccumulator meansAccumulator;

	public GRGENPlanGraphBuilder(MLSDMMetadataIndex metadataAdapter) {
		this(metadataAdapter, new GRGENGeometricMeansAccumulator(metadataAdapter));
	}

	public GRGENPlanGraphBuilder(MLSDMMetadataIndex metadataAdapter,
			GRGENGeometricMeansAccumulator meansAccumulator) {
		this.metadataAdapter = metadataAdapter;
		this.meansAccumulator = meansAccumulator;
	}

	public GRGENPlanGraph buildPlanGraph(SearchModel<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> searchModel) {
		GRGENPlanGraph planGraph = new GRGENPlanGraph();
		planGraph.root = new GRGENPlanGraphNode("root", searchModel);
		planGraph.nodes.add(planGraph.root);
		
		Map<Object, GRGENPlanGraphNode> nodeMap = new HashMap<Object, GRGENPlanGraphNode>();
		for(PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> node:searchModel.getPatternNodes()) {
			GRGENPlanGraphNode planGraphNode = new GRGENPlanGraphNode(node.getSpo().getName(), node);
			planGraph.nodes.add(planGraphNode);
			nodeMap.put(node, planGraphNode);
		}
		
		for(PatternConstraint<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> constraint:searchModel.getPatternConstraints()) {
			if(constraint.getConstraint() instanceof StoryPatternLink) {
				StoryPatternLink link = (StoryPatternLink) constraint.getConstraint();
				GRGENPlanGraphNode planGraphNode = new GRGENPlanGraphNode(link.getSource().getName() + "->" + link.getTarget().getName(), constraint);
				planGraph.nodes.add(planGraphNode);
				nodeMap.put(constraint, planGraphNode);
				createPlanGraphEdges(constraint, planGraph, nodeMap);
			}
			else {
				createPlanGraphEdges(constraint, planGraph, nodeMap);
			}
		}
		return planGraph;
	}
	
	private void createPlanGraphEdges(PatternConstraint<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> constraint, GRGENPlanGraph planGraph,
			Map<Object, GRGENPlanGraphNode> nodeMap) {
		Set<GRGENPlanGraphNode> toNodes = new HashSet<GRGENPlanGraphNode>();
		for(MatchingAction<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> action:constraint.getPossibleActions()) {
			GRGENPlanGraphEdge actionEdge = null;
			if(action instanceof MLSDMLinkLookupMatchingAction) {
				GRGENPlanGraphNode source = nodeMap.get(((MLSDMLinkLookupMatchingAction) action).getSourceNode());
				GRGENPlanGraphNode target = nodeMap.get(((MLSDMLinkLookupMatchingAction) action).getTargetNode());
				GRGENPlanGraphNode edge = nodeMap.get(constraint);
				
				actionEdge = new GRGENPlanGraphEdge(planGraph.root, edge, estimateCost(action));
				planGraph.edges.add(actionEdge);

				if(!toNodes.contains(source)) {
					GRGENPlanGraphEdge toSourceEdge = new GRGENPlanGraphEdge(edge, source, 0);
					planGraph.edges.add(toSourceEdge);
					toNodes.add(source);
				}
				if(!toNodes.contains(target)) {
					GRGENPlanGraphEdge toTargetEdge = new GRGENPlanGraphEdge(edge, target, 0);
					planGraph.edges.add(toTargetEdge);
					toNodes.add(target);
				}
			}
			else if(action instanceof MLSDMLinkTraverseMatchingAction) {
				GRGENPlanGraphNode source = nodeMap.get(((MLSDMLinkTraverseMatchingAction) action).getSourceNode());
				GRGENPlanGraphNode target = nodeMap.get(((MLSDMLinkTraverseMatchingAction) action).getTargetNode());
				GRGENPlanGraphNode edge = nodeMap.get(constraint);
				
				actionEdge = new GRGENPlanGraphEdge(source, edge, estimateCost(action));
				planGraph.edges.add(actionEdge);
				
				if(!toNodes.contains(target)) {
					GRGENPlanGraphEdge toTargetEdge = new GRGENPlanGraphEdge(edge, target, 0);
					planGraph.edges.add(toTargetEdge);
					toNodes.add(target);
				}
			}
			else if(action instanceof MLSDMLinkTraverseReverseMatchingAction) {
				GRGENPlanGraphNode source = nodeMap.get(((MLSDMLinkTraverseReverseMatchingAction) action).getSourceNode());
				GRGENPlanGraphNode target = nodeMap.get(((MLSDMLinkTraverseReverseMatchingAction) action).getTargetNode());
				GRGENPlanGraphNode edge = nodeMap.get(constraint);
				
				actionEdge = new GRGENPlanGraphEdge(target, edge, estimateCost(action));
				planGraph.edges.add(actionEdge);
				
				if(!toNodes.contains(source)) {
					GRGENPlanGraphEdge toSourceEdge = new GRGENPlanGraphEdge(edge, source, 0);
					planGraph.edges.add(toSourceEdge);
					toNodes.add(source);
				}
			}
			else if(action instanceof MLSDMDomainMatchMatchingAction) {
				GRGENPlanGraphNode node = nodeMap.get(((MLSDMDomainMatchMatchingAction) action).getNode());
				
				actionEdge = new GRGENPlanGraphEdge(planGraph.root, node, estimateCost(action));
				planGraph.edges.add(actionEdge);
			}
			
			if(actionEdge != null) {
				actionEdge.action = action;
			}
		}
	}

	private double estimateCost(
			MatchingAction<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> action) {
		return Math.log(Math.max(1, getRawEstimate(action)));
	}

	protected double getRawEstimate(
			MatchingAction<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> action) {
		if(action instanceof MLSDMLinkLookupMatchingAction) {
			return metadataAdapter.getReferences((EReference) ((MLSDMLinkLookupMatchingAction) action).getFeature()).size();
		}
		else if(action instanceof MLSDMLinkTraverseMatchingAction) {
			EReference reference = (EReference) ((MLSDMLinkTraverseMatchingAction) action).getFeature();
			return meansAccumulator.getForwardCost(reference);
		}
		else if(action instanceof MLSDMLinkTraverseReverseMatchingAction) {
			EReference reference = (EReference) ((MLSDMLinkTraverseReverseMatchingAction) action).getFeature();
			return meansAccumulator.getBackwardCost(reference);
		}
		else if(action instanceof MLSDMDomainMatchMatchingAction) {
			return (double) metadataAdapter.getDomain((EClass) action.getPatternConstraint().getConstraint()).size();
		}
		return 1;
	}
	

}
