package de.mdelab.mlsdm.interpreter.searchModel.patternMatcher.strategy.grgen;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Comparator;
import java.util.HashMap;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.PriorityQueue;
import java.util.Set;

import org.eclipse.emf.ecore.EClassifier;
import org.eclipse.emf.ecore.EStructuralFeature;

import de.mdelab.mlexpressions.MLExpression;
import de.mdelab.mlsdm.interpreter.searchModel.model.action.MLSDMDomainMatchMatchingAction;
import de.mdelab.mlsdm.interpreter.searchModel.model.action.MLSDMLinkLookupMatchingAction;
import de.mdelab.mlsdm.interpreter.searchModel.model.action.MLSDMLinkTraverseMatchingAction;
import de.mdelab.mlsdm.interpreter.searchModel.model.action.MLSDMLinkTraverseReverseMatchingAction;
import de.mdelab.mlsdm.interpreter.searchModel.patternMatcher.MLSDMMetadataIndex;
import de.mdelab.mlsdm.interpreter.searchModel.patternMatcher.strategy.hybrid.MLSDMStaticSelectionStrategy;
import de.mdelab.mlstorypatterns.AbstractStoryPatternLink;
import de.mdelab.mlstorypatterns.AbstractStoryPatternObject;
import de.mdelab.mlstorypatterns.StoryPattern;
import de.mdelab.mlsdm.interpreter.searchModel.patternMatcher.strategy.hybrid.MLSDMContinuationCostCalculator.CostModel;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.MatchingAction;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.PatternConstraint;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.PatternNode;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.SearchModel;
import de.mdelab.sdm.interpreter.core.variables.NotifierVariablesScope;

public class GRGENSelectionStrategy extends MLSDMStaticSelectionStrategy {

	private static int NODE_NAME_INDEX = 0;
	
	public GRGENSelectionStrategy(MLSDMMetadataIndex metadataAdapter,
			CostModel costModel,
			NotifierVariablesScope<?, ?, ?, ?, ?, ?, ?, ?, MLExpression> variablesScope) {
		super(metadataAdapter, costModel, variablesScope);
	}

	protected GRGENPlanGraphBuilder createPlanGraphBuilder() {
		return new GRGENPlanGraphBuilder(metadataAdapter);
	}
	
	@Override
	protected void doInitialize() {
		GRGENPlanGraphBuilder builder = createPlanGraphBuilder();
		GRGENPlanGraph planGraph = builder.buildPlanGraph(searchModel);
		
		List<MatchingAction<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>> actionOrder = computeMatchingOrder(planGraph);
		actionOrder = weaveChecks(actionOrder, searchModel);
		matchingOrder = new ArrayList<PatternConstraint<StoryPattern,AbstractStoryPatternObject,AbstractStoryPatternLink,EClassifier,EStructuralFeature,MLExpression>>();
		for(MatchingAction<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> action:actionOrder) {
			action.getPatternConstraint().setActiveAction(action);
			matchingOrder.add(action.getPatternConstraint());
		}
		matchingOrderInitialized = true;
	}

	protected List<MatchingAction<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>> weaveChecks(
			List<MatchingAction<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>> actionOrder,
			SearchModel<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> searchModel) {
		Set<PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>> boundNodes =
				new LinkedHashSet<PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>>();
		Set<PatternConstraint<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>> checkedConstraints =
				new LinkedHashSet<PatternConstraint<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>>();
		List<MatchingAction<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>> weavedOrder =
				new ArrayList<MatchingAction<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>>();
		
		for(MatchingAction<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> action:actionOrder) {
			weavedOrder.add(action);
			checkedConstraints.add(action.getPatternConstraint());
			
			Set<PatternConstraint<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>> candidateConstraints =
					new LinkedHashSet<PatternConstraint<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>>();
			for(PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> patternNode:action.getPatternConstraint().getDependencies()) {
				if(!boundNodes.contains(patternNode)) {
					boundNodes.add(patternNode);
					for(PatternConstraint<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> patternConstraint:patternNode.getDependantPatternConstraints()) {
						if(!checkedConstraints.contains(patternConstraint)) {
							candidateConstraints.add(patternConstraint);
						}
					}
				}
			}
			
			for(PatternConstraint<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> candidateConstraint:candidateConstraints) {
				if(boundNodes.containsAll(candidateConstraint.getDependencies())) {
					weavedOrder.add(candidateConstraint.getExplicitCheckAction());
					checkedConstraints.add(candidateConstraint);
				}
			}
		}
		
		return weavedOrder;
	}

	protected List<MatchingAction<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>> computeMatchingOrder(
			GRGENPlanGraph planGraph) {
		Collection<GRGENPlanGraphEdge> msa = computeMSA(planGraph);
		List<MatchingAction<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>> actionOrder = computeActionOrder(planGraph, msa);
		return actionOrder;
	}

	private List<MatchingAction<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>> computeActionOrder(
			GRGENPlanGraph planGraph, Collection<GRGENPlanGraphEdge> msa) {
		List<MatchingAction<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>> actionOrder =
				new ArrayList<MatchingAction<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>>();
		PriorityQueue<GRGENPlanGraphEdge> edgeQueue = new PriorityQueue<GRGENPlanGraphEdge>(new Comparator<GRGENPlanGraphEdge>() {

			@Override
			public int compare(GRGENPlanGraphEdge o1, GRGENPlanGraphEdge o2) {
				return Double.compare(o1.cost, o2.cost);
			}
			
		});
		Set<Object> reachedNodes = new LinkedHashSet<Object>();
		
		edgeQueue.addAll(getMSAEdges(planGraph.root.out, msa));
		while(!edgeQueue.isEmpty()) {
			GRGENPlanGraphEdge edge = edgeQueue.poll();
			
			if(edge.action != null) {
				if(edge.action instanceof MLSDMLinkTraverseMatchingAction) {
					reachedNodes.add(((MLSDMLinkTraverseMatchingAction) edge.action).getSourceNode());
					reachedNodes.add(((MLSDMLinkTraverseMatchingAction) edge.action).getTargetNode());
				}
				else if(edge.action instanceof MLSDMLinkTraverseReverseMatchingAction) {
					reachedNodes.add(((MLSDMLinkTraverseReverseMatchingAction) edge.action).getSourceNode());
					reachedNodes.add(((MLSDMLinkTraverseReverseMatchingAction) edge.action).getTargetNode());
				}
				else if(edge.action instanceof MLSDMLinkLookupMatchingAction) {
					reachedNodes.add(((MLSDMLinkLookupMatchingAction) edge.action).getSourceNode());
					reachedNodes.add(((MLSDMLinkLookupMatchingAction) edge.action).getTargetNode());
				}
				else if(edge.action instanceof MLSDMDomainMatchMatchingAction && reachedNodes.contains(((MLSDMDomainMatchMatchingAction)edge.action).getNode())) {
					continue;
				}
				
				actionOrder.add(edge.action);
			}
			
			edgeQueue.addAll(getMSAEdges(edge.target.out, msa));
		}
		return actionOrder;
	}

	private Collection<? extends GRGENPlanGraphEdge> getMSAEdges(
			List<GRGENPlanGraphEdge> edges, Collection<GRGENPlanGraphEdge> msa) {
		List<GRGENPlanGraphEdge> selectedEdges = new ArrayList<GRGENPlanGraphEdge>(edges.size());
		for(GRGENPlanGraphEdge edge:edges) {
			if(msa.contains(edge)) {
				selectedEdges.add(edge);
			}
		}
		return selectedEdges;
	}

	private Collection<GRGENPlanGraphEdge> computeMSA(GRGENPlanGraph planGraph) {
		Map<Object, Object> cloneMappings = planGraph.cloneWithMapping();
		Collection<GRGENPlanGraphEdge> cloneMSA = computeMSAEdmonds((GRGENPlanGraph) cloneMappings.get(planGraph));
		Set<GRGENPlanGraphEdge> msa = new LinkedHashSet<GRGENPlanGraphEdge>();
		for(GRGENPlanGraphEdge cloneEdge:cloneMSA) {
			msa.add((GRGENPlanGraphEdge) cloneMappings.get(cloneEdge));
		}
		return msa;
	}

	private Collection<GRGENPlanGraphEdge> computeMSAEdmonds(
			GRGENPlanGraph planGraph) {
		//remove edges to root
		for(int i = 0; i < planGraph.edges.size(); i++) {
			GRGENPlanGraphEdge edge = planGraph.edges.get(0);
			if(edge.target == planGraph.root) {
				planGraph.edges.remove(i);
				i--;
			}
		}
		
		List<GRGENPlanGraphEdge> candidate = computeCandidateMSA(planGraph);
		List<GRGENPlanGraphEdge> cycle = findCycle(planGraph, candidate);
		if(cycle == null) {
			return candidate;
		}
		else {
			GRGENPlanGraphNode compacted = new GRGENPlanGraphNode("compacted" + NODE_NAME_INDEX++, null);
			planGraph.nodes.add(compacted);
			
			Set<GRGENPlanGraphNode> cycleNodes = new LinkedHashSet<GRGENPlanGraphNode>();
			for(GRGENPlanGraphEdge edge:cycle) {
				cycleNodes.add(edge.source);
				cycleNodes.add(edge.target);
			}
			
			Map<GRGENPlanGraphEdge, GRGENPlanGraphEdge> edgeCorrespondences = new HashMap<GRGENPlanGraphEdge, GRGENPlanGraphEdge>();
			for(GRGENPlanGraphNode node:cycleNodes) {
				planGraph.nodes.remove(node);
				
				//handle edges from and to cycle
				for(GRGENPlanGraphEdge inEdge:node.in) {
					planGraph.edges.remove(inEdge);
					inEdge.source.out.remove(inEdge);
					if(!cycleNodes.contains(inEdge.source)) {
						GRGENPlanGraphEdge newEdge = new GRGENPlanGraphEdge(inEdge.source, compacted, inEdge.cost - getIncomingEdgeInCandidate(inEdge.target, candidate).cost);	//TODO weight!
						planGraph.edges.add(newEdge);
						edgeCorrespondences.put(newEdge, inEdge);
					}
				}
				for(GRGENPlanGraphEdge outEdge:node.out) {
					planGraph.edges.remove(outEdge);
					outEdge.target.in.remove(outEdge);
					if(!cycleNodes.contains(outEdge.target)) {
						GRGENPlanGraphEdge newEdge = new GRGENPlanGraphEdge(compacted, outEdge.target, outEdge.cost);
						planGraph.edges.add(newEdge);
						edgeCorrespondences.put(newEdge, outEdge);
					}
				}
			}
			
			if(cycleNodes.contains(planGraph.root)) {
				planGraph.root = compacted;
			}
			
			Collection<GRGENPlanGraphEdge> msaPrime = computeMSAEdmonds(planGraph);
			GRGENPlanGraphEdge inCompacted = null;
			for(GRGENPlanGraphEdge edge:msaPrime) {
				if(edge.target == compacted) {
					inCompacted = edgeCorrespondences.containsKey(edge) ? edgeCorrespondences.get(edge) : edge;
					break;
				}
			}
			
			List<GRGENPlanGraphEdge> msa = new ArrayList<GRGENPlanGraphEdge>();
			for(GRGENPlanGraphEdge cycleEdge:cycle) {
				if(cycleEdge.target != inCompacted.target) {
					msa.add(cycleEdge);
				}
			}
			for(GRGENPlanGraphEdge edge:msaPrime) {
				msa.add(edgeCorrespondences.containsKey(edge) ? edgeCorrespondences.get(edge) : edge);
			}
			
			return msa;
		}
	}

	private GRGENPlanGraphEdge getIncomingEdgeInCandidate(GRGENPlanGraphNode target, List<GRGENPlanGraphEdge> candidate) {
		for(GRGENPlanGraphEdge edge:candidate) {
			if(edge.target == target) {
				return edge;
			}
		}
		return null;
	}

	private List<GRGENPlanGraphEdge> findCycle(
			GRGENPlanGraph planGraph, Collection<GRGENPlanGraphEdge> candidate) {
		Set<GRGENPlanGraphNode> marking = new LinkedHashSet<GRGENPlanGraphNode>();
		List<GRGENPlanGraphEdge> cycle = null;
		for(GRGENPlanGraphNode node:planGraph.nodes) {
			if(marking.contains(node)) {
				continue;
			}
			
			marking.add(node);
			Set<GRGENPlanGraphNode> onPath = new LinkedHashSet<GRGENPlanGraphNode>();
			onPath.add(node);
			cycle = findCycleWithRoot(node, planGraph, candidate, marking, new ArrayList<GRGENPlanGraphEdge>(), onPath);
			if(cycle != null) { 
				return cycle;
			}
		}
		return null;
	}

	private List<GRGENPlanGraphEdge> findCycleWithRoot(GRGENPlanGraphNode root,
			GRGENPlanGraph planGraph, Collection<GRGENPlanGraphEdge> candidate,
			Set<GRGENPlanGraphNode> marking,
			ArrayList<GRGENPlanGraphEdge> currentPath,
			Set<GRGENPlanGraphNode> onPath) {
		for(GRGENPlanGraphEdge edge:root.out) {
			if(!candidate.contains(edge)) {
				continue;
			}
			//check if target has already been reached before
			if(onPath.contains(edge.target)) {
				List<GRGENPlanGraphEdge> cycle = new ArrayList<GRGENPlanGraphEdge>();
				int i;
				for(i = 0; i < currentPath.size() && currentPath.get(i).source != edge.target;) {
					i++;
				}
				for(; i < currentPath.size(); i++) {
					cycle.add(currentPath.get(i));
				}
				cycle.add(edge);
				return cycle;
			}
			else if(marking.contains(edge.target)) {
				continue;
			}

			currentPath.add(edge);
			marking.add(edge.target);
			onPath.add(edge.target);
			
			List<GRGENPlanGraphEdge> cycle = findCycleWithRoot(edge.target, planGraph, candidate, marking, currentPath, onPath);
			if(cycle != null) {
				return cycle;
			}
			
			onPath.remove(edge.target);
			currentPath.remove(currentPath.size() - 1);
		}
		return null;
	}

	private List<GRGENPlanGraphEdge> computeCandidateMSA(
			GRGENPlanGraph planGraph) {
		List<GRGENPlanGraphEdge> candidate = new ArrayList<GRGENPlanGraphEdge>();
		for(GRGENPlanGraphNode node:planGraph.nodes) {
			if(node == planGraph.root) {
				continue;
			}

			GRGENPlanGraphEdge minEdge = null;
			double minCost = Double.MAX_VALUE;
			for(GRGENPlanGraphEdge edge:node.in) {
				if(edge.cost < minCost) {
					minEdge = edge;
					minCost = edge.cost;
				}
			}
			if(minEdge != null) {
				candidate.add(minEdge);
			}
		}
		return candidate;
	}
	
}
