package de.mdelab.mlsdm.interpreter.searchModel.patternMatcher.strategy.hybrid;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.LinkedHashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.Map;

import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.EClassifier;
import org.eclipse.emf.ecore.EReference;
import org.eclipse.emf.ecore.EStructuralFeature;

import de.mdelab.mlexpressions.MLExpression;
import de.mdelab.mlsdm.interpreter.searchModel.model.action.MLSDMExpressionCheckMatchingAction;
import de.mdelab.mlsdm.interpreter.searchModel.model.action.MLSDMLinkCheckMatchingAction;
import de.mdelab.mlsdm.interpreter.searchModel.model.action.MLSDMSPOExpressionCheckMatchingAction;
import de.mdelab.mlsdm.interpreter.searchModel.patternMatcher.MLSDMMetadataIndex;
import de.mdelab.mlsdm.mlindices.Index;
import de.mdelab.mlstorypatterns.AbstractStoryPatternLink;
import de.mdelab.mlstorypatterns.AbstractStoryPatternObject;
import de.mdelab.mlstorypatterns.StoryPattern;
import de.mdelab.mlstorypatterns.StoryPatternLink;
import de.mdelab.sdm.icl.iCL.ICLConstraint;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.CheckMatchingAction;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.MatchingAction;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.PatternConstraint;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.PatternNode;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.SearchModel;
import de.mdelab.sdm.interpreter.core.variables.NotifierVariablesScope;

public class MLSDMConstraintSensitiveCostCalculator
		extends
			MLSDMContinuationCostCalculator {
	
	private static final double DEFAULT_FILTERING = 1;
	
	private Map<ICLConstraint, Double> indexFilters;

	public MLSDMConstraintSensitiveCostCalculator(
			SearchModel<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> searchModel,
			MLSDMMetadataIndex metadataAdapter, CostModel costModel,
			NotifierVariablesScope<?, ?, ?, ?, ?, ?, ?, ?, MLExpression> variablesScope) {
		super(searchModel, metadataAdapter, costModel, variablesScope);
		this.indexFilters = new HashMap<ICLConstraint, Double>();
	}
	
	@Override
	public Adornment[] calculateCotinuationCosts() {
		Adornment[] table = new Adornment[(int) Math.pow(2, searchModel.getPatternNodes().size())];
		Adornment initial = new Adornment();
		initial.appliedFilters = getAppliedFilters(initial.bindings);
		table[(int) Math.pow(2, searchModel.getPatternNodes().size()) - 1] = initial;
		
		@SuppressWarnings("unchecked")
		Collection<Adornment>[] configurations = (Collection<Adornment>[]) new Collection[searchModel.getPatternNodes().size() + 1];
		for(int i = 0; i < configurations.length; i++) {
			configurations[i] = new LinkedHashSet<Adornment>();
		}
		configurations[configurations.length - 1].add(initial);
		
		for(int i = configurations.length - 1; i >= 0; i--) {
			Collection<Adornment> adornments = configurations[i];
			for(Adornment a:adornments) {
				Collection<Adornment> predecessors = computePredecessors(a);
				for(Adornment predecessor:predecessors) {
					int bindingCode = getBindingCode(predecessor);
					if(table[bindingCode] == null || table[bindingCode].totalCost > predecessor.totalCost) {
						int boundNodeNumber = getBoundNodeNumber(predecessor);
						configurations[boundNodeNumber].remove(table[bindingCode]);
						configurations[boundNodeNumber].add(predecessor);
						table[bindingCode] = predecessor;
					}
				}
			}
		}
		return table;
	}

	private double getAppliedFilters(boolean[] bindings) {
		double appliedFilters = 1;
		for(PatternConstraint<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> pc:searchModel.getPatternConstraints()) {
			if(isChecked(pc, bindings)) {
				appliedFilters *= getFiltering(pc.getExplicitCheckAction());
			}
		}
		return appliedFilters;
	}

	protected Adornment createDerivedAdornment(Adornment a, boolean[] derivedBindings,
			double cost, int newBindings, MatchingAction<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> ma) {
		Collection<PatternConstraint<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>> checkedConstraints = new ArrayList<PatternConstraint<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>>(a.unhandledConstraints);
		LinkedList<PatternConstraint<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>> matchingOrder = new LinkedList<PatternConstraint<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>>(a.matchingOrder);
		LinkedList<MatchingAction<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>> actions = new LinkedList<MatchingAction<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>>(a.actions);
		double filtering = 1;
		for(Iterator<PatternConstraint<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>> it = checkedConstraints.iterator(); it.hasNext();) {
			PatternConstraint<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> pc2 = it.next();
			if(!isChecked(pc2, derivedBindings)) {
				it.remove();
				if(pc2 != ma.getPatternConstraint()) {
					matchingOrder.add(0, pc2);
					actions.add(0, pc2.getExplicitCheckAction());
					filtering *= getFiltering(pc2.getExplicitCheckAction());
				}
			}
		}
		matchingOrder.add(0, ma.getPatternConstraint());
		actions.add(0, ma);
		
		double remainingCost = (newBindings * cost) + (a.totalCost * cost * filtering);
		double newFiltering = (filtering * getFiltering(ma.getPatternConstraint().getExplicitCheckAction()));
		double appliedFiltering = newFiltering != 0 ? a.appliedFilters / newFiltering : getAppliedFilters(derivedBindings);
		
		return new Adornment(derivedBindings, checkedConstraints, remainingCost, matchingOrder, actions, appliedFiltering);
	}

	@Override
	public double getFiltering(
			CheckMatchingAction<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> explicitCheckAction) {
		switch(costModel) {
			case AVERAGE_CASE :
				return getAverageFiltering(explicitCheckAction);
			case WORST_CASE :
				return getWorstFiltering(explicitCheckAction);
			case BEST_CASE :
				return getBestFiltering(explicitCheckAction);
			default :
				return getAverageFiltering(explicitCheckAction);
		}
	}

	private double getBestFiltering(
			CheckMatchingAction<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> explicitCheckAction) {
		return getAverageFiltering(explicitCheckAction) < 1 ? 0 : 1;
	}

	private double getWorstFiltering(
			CheckMatchingAction<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> explicitCheckAction) {
		return getAverageFiltering(explicitCheckAction) > 0 ? 1 : 0;
	}

	private double getAverageFiltering(
			CheckMatchingAction<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> explicitCheckAction) {
		if(explicitCheckAction instanceof MLSDMSPOExpressionCheckMatchingAction || explicitCheckAction instanceof MLSDMExpressionCheckMatchingAction) {
			int validInstances = metadataAdapter.getValidInstanceNumbers(explicitCheckAction.getPatternConstraint().getConstraint());
			if(validInstances != MLSDMMetadataIndex.EXPRESSION_NOT_REGISTERED) {
				double domainSize = metadataAdapter.getDomain((EClass) explicitCheckAction.getRequiredBindings().get(0).getSpo().getType()).size();
				return domainSize == 0 ? 0 : (double) validInstances / domainSize;
			}
			else if (explicitCheckAction.getPatternConstraint().getConstraint() instanceof ICLConstraint) {
				ICLConstraint constraint = (ICLConstraint) explicitCheckAction.getPatternConstraint().getConstraint();
				if(!indexFilters.containsKey(constraint)) {
					double domainSizes = 1;
					for(PatternNode<?, AbstractStoryPatternObject, ?, EClassifier, ?, ?> node:explicitCheckAction.getRequiredBindings()) {
						domainSizes *= metadataAdapter.getDomain((EClass) node.getSpo().getType()).size();
					}
					String indexId = constraint.getIndexID();
					Index index = (Index) variablesScope.getVariable(indexId).getValue();
					if(constraint.getOperation().getName().equals("CONTAINS")) {
						indexFilters.put(constraint, domainSizes == 0 ? 0 : index.size() / domainSizes);
					}
					else if(constraint.getOperation().getName().equals("EXCLUDES")) {
						indexFilters.put(constraint, domainSizes == 0 ? 0 : 1.0 - (index.size() / domainSizes));
					}
					else {
						indexFilters.put(constraint, DEFAULT_FILTERING);
					}
				}
				return indexFilters.get(constraint);
			}
			else {
				return DEFAULT_FILTERING;
			}
		}
		else if(explicitCheckAction instanceof MLSDMLinkCheckMatchingAction) {
			StoryPatternLink link = ((MLSDMLinkCheckMatchingAction) explicitCheckAction).getLink();
			double domainSize = ((double) metadataAdapter.getDomain((EClass) link.getSource().getType()).size() * (double) metadataAdapter.getDomain((EClass) link.getTarget().getType()).size());
			return domainSize == 0 ? 0 : metadataAdapter.getReferences((EReference) link.getFeature()).size() / domainSize;	//TODO polymorphism?
		}
		else {
			return 1;
		}
	}

}
