package de.mdelab.mlsdm.interpreter.searchModel.patternMatcher.strategy.hybrid;

import org.eclipse.emf.ecore.EClassifier;
import org.eclipse.emf.ecore.EStructuralFeature;

import de.mdelab.mlexpressions.MLExpression;
import de.mdelab.mlsdm.interpreter.searchModel.patternMatcher.MLSDMMetadataIndex;
import de.mdelab.mlsdm.interpreter.searchModel.patternMatcher.strategy.MLSDMMatchingCostStrategy;
import de.mdelab.mlsdm.interpreter.searchModel.patternMatcher.strategy.hybrid.MLSDMContinuationCostCalculator.Adornment;
import de.mdelab.mlsdm.interpreter.searchModel.patternMatcher.strategy.hybrid.MLSDMContinuationCostCalculator.CostModel;
import de.mdelab.mlstorypatterns.AbstractStoryPatternLink;
import de.mdelab.mlstorypatterns.AbstractStoryPatternObject;
import de.mdelab.mlstorypatterns.StoryPattern;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.PatternConstraint;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.PatternNode;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.SearchModel;
import de.mdelab.sdm.interpreter.core.variables.NotifierVariablesScope;

public class MLSDMHybridCostStrategy
		implements
			MLSDMMatchingCostStrategy {
	
	protected CostModel costModel;
	protected Adornment[] continuationTable;
	protected MLSDMContinuationCostCalculator costCalculator;
	protected SearchModel<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> searchModel;
	protected MLSDMHybridSelectionStrategy selectionStrategy;
	protected NotifierVariablesScope<?, ?, ?, ?, ?, ?, ?, ?, MLExpression> variablesScope;
	protected MLSDMMetadataIndex metadataAdapter;
	
	public MLSDMHybridCostStrategy(
			MLSDMHybridSelectionStrategy selectionStrategy,
			SearchModel<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> searchModel,
			CostModel costModel,
			NotifierVariablesScope<?, ?, ?, ?, ?, ?, ?, ?, MLExpression> variablesScope,
			MLSDMMetadataIndex metadataAdapter) {
		this.selectionStrategy = selectionStrategy;
		this.searchModel = searchModel;
		this.costModel = costModel;
		this.variablesScope = variablesScope;
		this.metadataAdapter = metadataAdapter;
	}

	public void setCostCalculator(MLSDMConstraintSensitiveCostCalculator costCalculator) {
		this.costCalculator = costCalculator;
	}
	
	public void setContinuationTable(Adornment[] continuationTable) {
		this.continuationTable = continuationTable;
	}
	
	public void setSearchModel(SearchModel<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> searchModel) {
		this.searchModel = searchModel;
	}
	
	@Override
	public double calculateMatchingCost(
			PatternConstraint<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> patternConstraint) {
		if(patternConstraint.getCardinality() == PatternConstraint.CHECK_ONLY) {
			return PatternConstraint.CHECK_ONLY;
		}
		
		int oldBindingCode = selectionStrategy.getCurrentBinding();
		int newBindingCode = oldBindingCode;
		int newBindings = 0;
		for(PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> node:patternConstraint.getDependencies()) {
			if(!node.isBound()) {
				newBindingCode = newBindingCode | (1 << selectionStrategy.getIndex(node));
				newBindings++;
			}
		}
		
		double constraintCardinality = patternConstraint.getCardinality();
		double oldFiltering = continuationTable[oldBindingCode].appliedFilters;
		double newFiltering = continuationTable[newBindingCode].appliedFilters;
		double constraintFiltering = costCalculator.getFiltering(patternConstraint.getExplicitCheckAction());
		double filtering = (constraintFiltering == 0 || oldFiltering == 0) ? 0 : newFiltering / (constraintFiltering * oldFiltering);
		return 	(constraintCardinality * newBindings) +
				(constraintCardinality * filtering * continuationTable[newBindingCode].totalCost);
	}

	@Override
	public void initialize() {
		costCalculator = new MLSDMConstraintSensitiveCostCalculator(searchModel, metadataAdapter, costModel, variablesScope);
		continuationTable = costCalculator.calculateCotinuationCosts();
	}

}
