package de.mdelab.mlsdm.interpreter.searchModel.patternMatcher.strategy.hybrid;

import java.util.Collection;
import java.util.LinkedList;

import org.eclipse.emf.ecore.EClassifier;
import org.eclipse.emf.ecore.EStructuralFeature;

import de.mdelab.mlexpressions.MLExpression;
import de.mdelab.mlsdm.interpreter.searchModel.patternMatcher.MLSDMMetadataIndex;
import de.mdelab.mlsdm.interpreter.searchModel.patternMatcher.strategy.hybrid.MLSDMContinuationCostCalculator.Adornment;
import de.mdelab.mlsdm.interpreter.searchModel.patternMatcher.strategy.hybrid.MLSDMContinuationCostCalculator.CostModel;
import de.mdelab.mlstorypatterns.AbstractStoryPatternLink;
import de.mdelab.mlstorypatterns.AbstractStoryPatternObject;
import de.mdelab.mlstorypatterns.StoryPattern;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.PatternConstraint;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.PatternNode;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.strategy.OrderProducingSelectionStrategy;
import de.mdelab.sdm.interpreter.core.variables.NotifierVariablesScope;

public class MLSDMRevisedHybridSelectionStrategy 
		extends 
			OrderProducingSelectionStrategy<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> {

	protected MLSDMMetadataIndex metadataAdapter;
	protected CostModel costModel;
	protected NotifierVariablesScope<?, ?, ?, ?, ?, ?, ?, ?, MLExpression> variablesScope;
	
	protected int currentBinding;
	protected LinkedList<Collection<PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>>> bindingsStack;
	private Adornment[] continuationTable;
	private MLSDMConstraintSensitiveCostCalculator costCalculator;
	
	public MLSDMRevisedHybridSelectionStrategy(MLSDMMetadataIndex metadataAdapter, CostModel costModel,
										NotifierVariablesScope<?, ?, ?, ?, ?, ?, ?, ?, MLExpression> variablesScope) {
		super();
		this.metadataAdapter = metadataAdapter;
		this.costModel = costModel;
		this.variablesScope = variablesScope;
		this.currentBinding = 0;
		this.bindingsStack = new LinkedList<Collection<PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>>>();
	}
	
	@Override
	protected void doInitialize() {
		costCalculator = new MLSDMConstraintSensitiveCostCalculator(searchModel, metadataAdapter, costModel, variablesScope);
		continuationTable = costCalculator.calculateCotinuationCosts();
	}
	
	@Override
	protected void doUpdate(Collection<PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>> affectedNodes) {
		for(PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> node:affectedNodes) {
			addBinding(node);
		}
		bindingsStack.push(affectedNodes);
	}

	@Override
	protected void doRollBackLastUpdate() {
		Collection<PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>> affectedNodes = bindingsStack.pop();
		for(PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> node:affectedNodes) {
			removeBinding(node);
		}
	}
	
	protected void addBinding(
			PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> node) {
		int index = getIndex(node);
		currentBinding = currentBinding | (1 << index);
	}
	
	protected void removeBinding(
			PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> node) {
		int index = getIndex(node);
		currentBinding = currentBinding & (~(1 << index));
	}

	protected int getIndex(
			PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> node) {
		return node.getId();
	}

	public int getCurrentBinding() {
		return currentBinding;
	}
	
	@Override
	protected PatternConstraint<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> selectNextConstraint() {
		PatternConstraint<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> cheapestConstraint = getCheapestConstraint(false);
		return cheapestConstraint != null ? cheapestConstraint : getCheapestConstraint(true);
	}

	private PatternConstraint<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> getCheapestConstraint(
			boolean optional) {
		/*
		 * TODO problematic if for pc1, pc2 with continuationCode(pc1) == continuationCode(pc2)
		 * there is pc3 with cost(pc1) < cost(pc3) < cost(pc2) und cardinality(pc2) < cardinality(pc1).
		 * Idea: use array to cache values for each constraint, id to mark whether entry recent?
		 */
		int cheapestContinuationCode = -1;
		double cheapestConstraintCardinality = PatternConstraint.MATCHING_NOT_POSSIBLE;
		PatternConstraint<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> cheapestConstraint = null;
		double minCost = PatternConstraint.MATCHING_NOT_POSSIBLE;
		
		for(PatternConstraint<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> constraint:searchModel.getPatternConstraints()) {
			if(constraint.isActive() 
					|| constraint.getCardinality() == PatternConstraint.MATCHING_NOT_POSSIBLE 
					|| constraint.isOptional() != optional) {
				continue;
			}
			else if(constraint.getCardinality() == PatternConstraint.CHECK_ONLY) {
				return constraint;
			}
			
			int oldBindingCode = getCurrentBinding();
			int newBindingCode = oldBindingCode;
			int newBindings = 0;
			for(PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> node:constraint.getDependencies()) {
				if(!node.isBound()) {
					newBindingCode = newBindingCode | (1 << getIndex(node));
					newBindings++;
				}
			}
			double constraintCardinality = constraint.getCardinality();
			
			
			double oldFiltering = continuationTable[oldBindingCode].appliedFilters;
			double newFiltering = continuationTable[newBindingCode].appliedFilters;
			double constraintFiltering = costCalculator.getFiltering(constraint.getExplicitCheckAction());
			double filtering = (constraintFiltering == 0 || oldFiltering == 0) ? 0 : newFiltering / (constraintFiltering * oldFiltering);
			double cost = (constraintCardinality * newBindings) +
					(constraintCardinality * filtering * continuationTable[newBindingCode].totalCost);
		
			if(newBindingCode == cheapestContinuationCode) {
				if(constraintCardinality < cheapestConstraintCardinality) {
					cheapestContinuationCode = newBindingCode;
					cheapestConstraintCardinality = constraintCardinality;
					cheapestConstraint = constraint;
					minCost = minCost < cost ? minCost : cost;
				}
			}
			else {
				if(cost < minCost) {
					cheapestContinuationCode = newBindingCode;
					cheapestConstraintCardinality = constraintCardinality;
					cheapestConstraint = constraint;
					minCost = cost;
				}
			}
		}
		
		return cheapestConstraint;
	}

}
