package de.mdelab.mlsdm.interpreter.searchModel.patternMatcher.strategy.hybrid;

import java.util.ArrayList;
import java.util.Collection;

import org.eclipse.emf.ecore.EClassifier;
import org.eclipse.emf.ecore.EStructuralFeature;

import de.mdelab.mlexpressions.MLExpression;
import de.mdelab.mlsdm.interpreter.searchModel.patternMatcher.MLSDMMetadataIndex;
import de.mdelab.mlsdm.interpreter.searchModel.patternMatcher.strategy.hybrid.MLSDMContinuationCostCalculator.Adornment;
import de.mdelab.mlsdm.interpreter.searchModel.patternMatcher.strategy.hybrid.MLSDMContinuationCostCalculator.CostModel;
import de.mdelab.mlstorypatterns.AbstractStoryPatternLink;
import de.mdelab.mlstorypatterns.AbstractStoryPatternObject;
import de.mdelab.mlstorypatterns.MatchTypeEnum;
import de.mdelab.mlstorypatterns.StoryPattern;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.MatchingAction;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.PatternConstraint;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.PatternNode;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.strategy.OrderProducingSelectionStrategy;
import de.mdelab.sdm.interpreter.core.variables.NotifierVariablesScope;

public class MLSDMStaticSelectionStrategy
		extends
			OrderProducingSelectionStrategy<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> {

	protected MLSDMMetadataIndex metadataAdapter;
	protected CostModel costModel;
	protected int matchingStep;
	protected NotifierVariablesScope<?, ?, ?, ?, ?, ?, ?, ?, MLExpression> variablesScope;
	protected Adornment[] table;
	protected boolean matchingOrderInitialized;
	
	public MLSDMStaticSelectionStrategy(MLSDMMetadataIndex metadataAdapter, CostModel costModel,
										NotifierVariablesScope<?, ?, ?, ?, ?, ?, ?, ?, MLExpression> variablesScope) {
		super();
		this.metadataAdapter = metadataAdapter;
		this.costModel = costModel;
		this.matchingStep = 0;
		this.variablesScope = variablesScope;
		this.matchingOrderInitialized = false;
	}

	@Override
	protected void doInitialize() {
		MLSDMContinuationCostCalculator c = new MLSDMContinuationCostCalculator(searchModel, metadataAdapter, costModel, variablesScope);
		table = c.calculateCotinuationCosts();
	}
	
	@Override
	protected PatternConstraint<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> doPopPatternConstraint() {
		if(!matchingOrderInitialized) {
			int bindingCode = 0;
			for(int i = 0; i < searchModel.getPatternNodes().size(); i++) {
				PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> node
						= searchModel.getPatternNodes().get(i);
				if(node.isBound() && !(node.getSpo().getMatchType() == MatchTypeEnum.OPTIONAL)) {
					bindingCode = bindingCode | (1 << i);
				}
			}
			matchingOrder = new ArrayList<PatternConstraint<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>>(table[bindingCode].matchingOrder);
			System.out.println(matchingOrder);
			for(MatchingAction<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> ma:table[bindingCode].actions) {
				ma.getPatternConstraint().setActiveAction(ma);
			}
			matchingOrderInitialized = true;
		}
		PatternConstraint<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> constraint = super.doPopPatternConstraint();
		matchingStep++;
		return constraint;
	}
	
	@Override
	protected PatternConstraint<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> doRollBackLastPop() {
		matchingStep--;
		return super.doRollBackLastPop();
	}

	@Override
	protected void doUpdate(
			Collection<PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>> newlyBoundPatternNodes) {
	
	}

	@Override
	protected void updateSearchModel(
			Collection<PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>> newlyBoundPatternNodes,
			boolean updateMatchingCosts) {
		super.updateSearchModel(newlyBoundPatternNodes, false);
	}
	
	@Override
	protected void doRollBackLastUpdate() {
		
	}

	@Override
	protected PatternConstraint<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> selectNextConstraint() {
		return matchingOrder.get(matchingStep);
	}

	@Override
	public void reset() {
		super.reset();
		matchingStep = 0;
		matchingOrderInitialized = false;
	}
}
